# Defaults
# --------

SKIPLIST_DIR ?= skip
INCLIST_DIR ?= inc
STAGE1_LOG ?= logs/stage1.log
STAGE1_STATUS ?= stage1.status
STAGE1_LOGDIR ?= logs/stage1.logs

TAR_DIR ?= pkgs
STAGE2_LOG ?= logs/stage2.log
STAGE2_STATUS ?= stage2.status
STAGE2_LOGDIR ?= logs/stage2.logs

SRPM_DIR ?= out
STAGE3_LOG ?= logs/stage3.log
STAGE3_STATUS ?= stage3.status
STAGE3_LOGDIR ?= logs/stage3.logs

RPM_DIR ?= rpms
STAGE4_LOG ?= logs/stage4.log
STAGE4_STATUS ?= stage4.status
STAGE4_LOGDIR ?= logs/stage4.logs

REBUILD_V_DIR ?= rebuild.v
STAGE5_V_LOG ?= logs/stage5.v.log
STAGE5_V_STATUS ?= stage5.v.status
STAGE5_V_LOGDIR ?= logs/stage5.v.logs
REBUILD_O_DIR ?= rebuild.o
STAGE5_O_LOG ?= logs/stage5.o.log
STAGE5_O_STATUS ?= stage5.o.status
STAGE5_O_LOGDIR ?= logs/stage5.o.logs

REPO_FILES_SUBDIR ?= /files
REPO_SRPM_SUBDIR ?= $(REPO_FILES_SUBDIR)/SRPMS
USER ?= builder
TMPDIR ?= /tmp/.private/$(USER)
HASHER_WORKDIR ?= $(TMPDIR)/hashers

REPO_DIR ?= /repo/sisyphus
BIN_LIST ?= $(REPO_DIR)$(REPO_FILES_SUBDIR)/list/bin.list
DISTTAG_INDEX ?= disttag.index
DISTTAG_OPTS ?= --disttag-index="$(DISTTAG_INDEX)"

EXPORT_OPTS ?=
OUTREPO ?= repo
EXPORT_REPO_OPTS ?= $(EXPORT_OPTS)

INDEP_LOGS ?= 1
TOOLERANCE ?= 0
SSHDELAY ?= 0.1


# Tools

HSH_SEPARATE_SOURCES ?= $(if $(wildcard ./hsh-separate-sources),$(wildcard ./hsh-separate-sources),hsh-separate-sources)

SRPM_CLEANUP ?= $(if $(wildcard ./srpm-cleanup),$(wildcard ./srpm-cleanup),srpm-cleanup)

IMPORT_LOGS ?= $(firstword $(wildcard ./import-logs.pl /usr/share/srpm-cleanup/import-logs.pl))

COMMON_OPTS ?= $(if $(SELFIE),--selfie) $(if $(CLEANUP),--cleanup) $(if $(REMOTE_DIR),--remote-dir="$(REMOTE_DIR)") $(if $(SSHDELAY),--sshdelay $(SSHDELAY)) $(if $(WITH_SRPMS),--with-srpms="$(WITH_SRPMS)") $(if $(LIMIT),--limit=$(LIMIT))

BUILD_COMMON_OPTS = $(COMMON_OPTS) $(if $(WITH_NPROFILE),--with-nprofile="$(WITH_NPROFILE)") $(if $(MAX_SILENCE),--max-silence=$(MAX_SILENCE))

CLEAN_COMMON_OPTS = $(COMMON_OPTS)

ALL_SRPMS ?= all.srpms

# --- END OF CONFIGURATION --- #

# Automatic SRPM lists

auto_srpm_list_arches := $(if $(AUTO_SRPM_LIST_DIR),$(shell find $(AUTO_SRPM_LIST_DIR) -type f -name '*.srpms' | sed -e 's/[[:space:]]\+/\n/g' | sed -n -e 's/^.*-\([^.-]\+\)\..*srpms$$/\1/p' | sort -u))
auto_srpm_list_dirs := $(if $(AUTO_SRPM_LIST_DIR),$(shell find $(AUTO_SRPM_LIST_DIR) -type d))

SRPM_LISTS += $(foreach dir,$(auto_srpm_list_dirs),$(foreach arch,$(auto_srpm_list_arches),$(call def_srpm_list,$(arch),$(firstword $(filter %-$(arch).srpms %-$(arch).iso.srpms,$(shell ls -t $(dir)/*.*srpms 2>/dev/null))))))

# Functions

def_srpm_list = $$(if $$(1),$$(if $$(filter $$(1),$(1)),$(2)),$(2))

define srpm_lists_template =
 SRPM_LISTS = $$(strip $(SRPM_LISTS))
endef
$(eval $(srpm_lists_template))

builder_cmd = -H "$(1)" -a "$(2)" -n "$(if $(SLOTS),$(SLOTS),$(3))" -d "$(4)" -R "$(5)" $$(if $$(1),,$$(if $$(WITH_SRPMS),,-B "$(5)$(REPO_SRPM_SUBDIR)")) $$(if $$(2),,$$(if $$(NPROCS),-j "$$(NPROCS)",$(if $(6),-j "$(6)")))
def_builder = $$$$(if $$$$(1),$$$$(if $$$$(filter $$$$(1),$(2)),$(call builder_cmd,$(1),$$$$(filter $$$$(1),$(2)),$(3),$(4),$(5),$(6))),$(call builder_cmd,$(1),$(2),$(3),$(4),$(5),$(6))) $(eval _ARCHES += $(2))

define builders_template1 =
 _BUILDERS = $(BUILDERS)
 BUILDERS = $$(_BUILDERS)
 STAGE4_BUILDERS = $$(call _BUILDERS,1)
 SRPM_BUILDERS = $$(call _BUILDERS,,1)
endef
$(eval $(builders_template1))

define builders_template2 =
 BUILDERS = $$(strip $(BUILDERS))
 STAGE4_BUILDERS = $$(strip $(STAGE4_BUILDERS))
 SRPM_BUILDERS = $$(strip $(SRPM_BUILDERS))
endef
$(eval $(builders_template2))

ARCHES = $(strip $(sort $(_ARCHES)) $(EXTRA_ARCHES))

cleaner_cmd = -H "$(1)" -n "$(if $(SLOTS),$(SLOTS),$(2))" -d "$(3)" $$(if $$(WITH_SRPMS),,-B "$(4)")
def_cleaner = $$(if $$(1),$$(if $$(filter $$(1),$(1)),$(call cleaner_cmd,$(1),$(2),$(3),$(4))),$(call cleaner_cmd,$(1),$(2),$(3),$(4)))

define cleaners_template =
 CLEANERS = $$(strip $(CLEANERS))
endef
$(eval $(cleaners_template))

MAKE_SELF = $(MAKE) --no-print-directory -f $(firstword $(MAKEFILE_LIST))

filter_status = sed -n -e 's/^---[0-9]\+\t\([^\t]\+\)\t[^\t]*\t$(2)\t.*$$/\1/p' $(1) | sort -u
filter_arch_status = sed -n -e 's/^---[0-9]\+\t\([^\t]\+\)\t$(2)\t$(3)\t.*$$/\1/p' $(1) | sort -u

backup_log = $(if $(DRY_RUN),,$(if $(BACKUP_LOGS),cp --backup=t /dev/null "$(1)",$(if $(INDEP_LOGS),$(eval LOG_SUF = $(shell mktemp "$(1).XXXX" | sed -e 's/^.*\././')))))
import_log = $(if $(DRY_RUN),,touch "$(2)"; flock "$(2)" $(IMPORT_LOGS) -o "$(2)" "$(2)" "$(1)")
import_log_noarch = $(if $(DRY_RUN),,touch "$(2)"; flock "$(2)" $(IMPORT_LOGS) -n -o "$(2)" "$(2)" "$(1)")

show_pkg_status = sed -n -e "/^---[0-9]\\+\\t$(2)\\t$(if $(3),$(3)\\t)/ { H; g; s/\\n//; p; z; h; d }" -e '/^---[0-9]/ { z; h; d }' -e 'H' "$(1)"

space := $(eval) $(eval)
comma = ,
backslash := $(strip \ )

# Rules

phonytargets =

status:
	@echo "--- Stage I ---"
	@$(MAKE_SELF) stage1-status
	@echo
	@echo "--- Stage II ---"
	@$(MAKE_SELF) stage2-status
	@echo
	@echo "--- Stage III ---"
	@$(MAKE_SELF) stage3-status
	@echo
	@echo "--- Stage IV ---"
	@$(MAKE_SELF) stage4-status
	@echo
	@echo "--- Stage V/1 ---"
	@$(MAKE_SELF) stage5v-status
	@echo
	@echo "--- Stage V/2 ---"
	@$(MAKE_SELF) stage5o-status
phonytargets += status

stage5-status:
	@echo "--- Stage V/1 ---"
	@$(MAKE_SELF) stage5v-status
	@echo
	@echo "--- Stage V/2 ---"
	@$(MAKE_SELF) stage5o-status
phonytargets += stage5-status

show-builders:
	@echo '$(BUILDERS)'
phonytargets += show-builders

show-cleaners:
	@echo '$(CLEANERS)'
phonytargets += show-cleaners

show-srpm-builders:
	@echo '$(SRPM_BUILDERS)'
phonytargets += show-srpm-builders

_always_:
phonytargets += _always_

$(ALL_SRPMS): $(if $(SRPM_LISTS),$(SRPM_LISTS),$(warning WARNING: No SRPM lists defined!) _always_)
	@$(if $(SRPM_LISTS),cat $(SRPM_LISTS) | sed -e '/^i586-/ d' -e '/^[[:space:]]*$$/ d' | sort -u,echo -n) >$@

list-srpms: $(ALL_SRPMS)
	@cat "$(ALL_SRPMS)"
phonytargets += list-srpms

define print_arch_srpms_template
 list-srpms-$(1): $$(if $$(call SRPM_LISTS,$(1)),$$(call SRPM_LISTS,$(1)),_always_)
	@$$(if $$(call SRPM_LISTS,$(1)),cat $$(call SRPM_LISTS,$(1)) | sed -e '/^i586-/ d' -e '/^[[:space:]]*$$$$/ d' | sort -u,echo -n)
 phonytargets += list-srpms-$(1)
endef
$(foreach arch,$(ARCHES),$(eval $(call print_arch_srpms_template,$(arch))))

list-arches:
	@echo "$(ARCHES)"
phonytargets += list-arches

define print_arch_builders_template =
 list-$(1)-builders:
	@echo "$$(call BUILDERS,$(1))"
 phonytargets += list-$(1)-builders
endef
$(foreach arch,$(ARCHES),$(eval $(call print_arch_builders_template,$(arch))))

package_lists =
define included_arch_list_template =
 .included.$(1).list: $$(if $$(call SRPM_LISTS,$(1)),$$(call SRPM_LISTS,$(1)),$$(warning WARNING: No SRPM list defined for $(1)!) _always_)
	@$$(if $$(call SRPM_LISTS,$(1)),cat $$(call SRPM_LISTS,$(1)) | sed -e '/^i586-/ d' -e '/^[[:space:]]*$$$$/ d' -e 's/\.src\.rpm//' | sort -u,echo -n) >$$@
 package_lists += .included.$(1).list
endef
$(foreach arch,$(ARCHES),$(eval $(call included_arch_list_template,$(arch))))

define united_list_template =
 $(1).list: $$(if $$(ARCHES),$$(foreach arch,$$(ARCHES),$(1).$$(arch).list),| _always_)
	@$$(if $$^,sort -u $$^,echo -n) >$$@
 package_lists += $(1).list
endef

$(eval $(call united_list_template,.included))

### --- Stage I ---

define stage1_unused_arch_list_template =
 .stage1-findlist.$(1).list: $$(SKIPLIST_DIR)
	@find $$(SKIPLIST_DIR) -name "*.$(1).skiplist" -printf '%f\n' | sed -e 's/\.[^.]\+\.skiplist$$$$//' | sort -u >$$@
 package_lists += .stage1-findlist.$(1).list
 .stage1-unused.$(1).list: .included.$(1).list .stage1-findlist.$(1).list
	@comm -13 ".included.$(1).list" ".stage1-findlist.$(1).list" >$$@
 package_lists += .stage1-unused.$(1).list
 clean-stage1-$(1)-unused: .stage1-unused.$(1).list
	@cat $$< | while read -r f; do rm -fv $$(RMOPTS) "$$(SKIPLIST_DIR)/$$$$f.$(1).skiplist"; done
 phonytargets += clean-stage1-$(1)-unused
endef
$(foreach arch,$(ARCHES),$(eval $(call stage1_unused_arch_list_template,$(arch))))

define stage1_arch_status_list_template =
 .stage1-$(1).$(2).list: $$(STAGE1_STATUS) .included.$(2).list
	@$$(call filter_arch_status,$$(STAGE1_STATUS),$(2),REBUILD\t$(3)) | comm -12 ".included.$(2).list" - >$$@
 package_lists += .stage1-$(1).$(2).list
endef
define stage1_status_list_template =
 $$(foreach arch,$$(ARCHES),$$(eval $$(call stage1_arch_status_list_template,$(1),$$(arch),$(2))))
endef

$(eval $(call stage1_status_list_template,statdone,\(LOGOK\|ALRDY\)))
$(eval $(call stage1_status_list_template,excl,BLDERR\tEXCL))
$(eval $(call stage1_status_list_template,blderr,BLDERR))
$(eval $(call stage1_status_list_template,brkndep,BLDERR\tBRKNDEP))
$(eval $(call stage1_status_list_template,chkerr,BLDERR\tCHKERR))
$(eval $(call stage1_status_list_template,nospace,BLDERR\tNOSPACE))
$(eval $(call stage1_status_list_template,noinode,BLDERR\tNOINODE))
$(eval $(call stage1_status_list_template,logerr,LOGERR))

define stage1_arch_lists_template =
 .stage1-other.$(1).list: .stage1-blderr.$(1).list .stage1-excl.$(1).list .stage1-brkndep.$(1).list .stage1-chkerr.$(1).list .stage1-nospace.$(1).list .stage1-noinode.$(1).list
	@cat ".stage1-excl.$(1).list" ".stage1-brkndep.$(1).list" ".stage1-chkerr.$(1).list" ".stage1-nospace.$(1).list" ".stage1-noinode.$(1).list" | sort -u | comm -23 ".stage1-blderr.$(1).list" - >$$@
 .stage1-noatt.$(1).list: .included.$(1).list .stage1-done.$(1).list .stage1-excl.$(1).list .stage1-blderr.$(1).list .stage1-logerr.$(1).list
	@cat ".stage1-done.$(1).list" ".stage1-excl.$(1).list" ".stage1-blderr.$(1).list" ".stage1-logerr.$(1).list" | sort -u | comm -23 ".included.$(1).list" - >$$@
 .stage1-done.$(1).list: .stage1-statdone.$(1).list .stage1-findlist.$(1).list
	@comm -12 ".stage1-statdone.$(1).list" ".stage1-findlist.$(1).list" >$$@
 package_lists += .stage1-other.$(1).list .stage1-noatt.$(1).list .stage1-done.$(1).list
endef
$(foreach arch,$(ARCHES),$(eval $(call stage1_arch_lists_template,$(arch))))

statcats1 = unused done excl blderr brkndep chkerr nospace noinode logerr other noatt

$(foreach cat,$(statcats1),$(eval $(call united_list_template,.stage1-$(cat))))

define print_list_template =
 list-$(1)-$(2): .$(1)-$(2).list
	@cat $$<
 phonytargets += list-$(1)-$(2)
 list-$(1)-$(2)-files: .$(1)-$(2).list
	@cat $$< | sed -e 's/^.*$$$$/&.src.rpm/' | sort -u
 phonytargets += list-$(1)-$(2)-files
 list-$(1)-$(2)-names: .$(1)-$(2).list
	@cat $$< | sed -e 's/-[^-]\+-[^-]\+$$$$//' | sort -u
 phonytargets += list-$(1)-$(2)-names
 show-$(1)-$(2)-status: .$(1)-$(2).list
	@cat $$< | while read -r pkg; do $$(call show_pkg_status,$$($(patsubst %:$(1),%,$(filter %:$(1),STAGE1_STATUS:stage1 STAGE2_STATUS:stage2 STAGE3_STATUS:stage3 STAGE4_STATUS:stage4 STAGE5_V_STATUS:stage5v STAGE5_O_STATUS:stage5o))),$$$$pkg); done
 phonytargets += show-$(1)-$(2)-status
endef
$(foreach cat,$(statcats1),$(eval $(call print_list_template,stage1,$(cat))))

define print_arch_list_template =
 list-$(1)-$(2)-$(3): .$(1)-$(2).$(3).list
	@cat $$<
 phonytargets += list-$(1)-$(2)-$(3)
 list-$(1)-$(2)-$(3)-files: .$(1)-$(2).$(3).list
	@cat $$< | sed -e 's/^.*$$$$/&.src.rpm/' | sort -u
 phonytargets += list-$(1)-$(2)-$(3)-files
 list-$(1)-$(2)-$(3)-names: .$(1)-$(2).$(3).list
	@cat $$< | sed -e 's/-[^-]\+-[^-]\+$$$$//' | sort -u
 phonytargets += list-$(1)-$(2)-$(3)-names
 show-$(1)-$(2)-$(3)-status: .$(1)-$(2).$(3).list
	@cat $$< | while read -r pkg; do $$(call show_pkg_status,$$($(patsubst %:$(1),%,$(filter %:$(1),STAGE1_STATUS:stage1 STAGE2_STATUS:stage2 STAGE3_STATUS:stage3 STAGE4_STATUS:stage4 STAGE5_V_STATUS:stage5v STAGE5_O_STATUS:stage5o))),$$$$pkg,$(3)); done
 phonytargets += show-$(1)-$(2)-$(3)-status
endef
$(foreach cat,$(statcats1),$(foreach arch,$(ARCHES),$(eval $(call print_arch_list_template,stage1,$(cat),$(arch)))))

define print_arch_logs_template =
 list-$(1)-$(2)-$(3)-logs: .$(1)-$(2).$(3).list
	@cat $$< | sed -e 's:^.*$$$$:$$($(patsubst %:$(1),%,$(filter %:$(1),STAGE1_LOGDIR:stage1 STAGE2_LOGDIR:stage2 STAGE3_LOGDIR:stage3 STAGE4_LOGDIR:stage4 STAGE5_V_LOGDIR:stage5v STAGE5_O_LOGDIR:stage5o)))/&.$(3).log:' | sort -u
 phonytargets += list-$(1)-$(2)-$(3)-logs
endef
$(foreach cat,$(statcats1),$(foreach arch,$(ARCHES),$(eval $(call print_arch_logs_template,stage1,$(cat),$(arch)))))

%-all.list: $(if $(ARCHES),$(foreach arch,$(ARCHES),%.$(arch).list),| _always_)
	@$(if $^,cat $(firstword $^) $(foreach f,$(wordlist 2,$(words $^),$^),| comm -12 $(f) -),echo -n) | sort -u >$@

.stage1-nosrpm.list: $(ALL_SRPMS) $(STAGE1_STATUS) .stage1-done-all.list .stage1-blderr-all.list .stage1-logerr-all.list
	@$(call filter_status,$(STAGE1_STATUS),REBUILD\tNOTFOUND) | comm -13 ".stage1-done-all.list" - | comm -13 ".stage1-blderr-all.list" - | comm -13 ".stage1-logerr-all.list" - | sed -e 's/^.*$$/&.src.rpm/' | sort -u | comm -12 "$(ALL_SRPMS)" - | sed -e 's/\.src\.rpm$$//' | sort -u >$@
package_lists += .stage1-done-all.list .stage1-blderr-all.list .stage1-logerr-all.list .stage1-nosrpm.list

list-stage1-nosrpm: .stage1-nosrpm.list
	@cat $<
phonytargets += list-stage1-nosrpm

stage1-status: $(ALL_SRPMS) .stage1-nosrpm.list $(foreach cat,$(statcats1),$(foreach arch,$(ARCHES),.stage1-$(cat).$(arch).list)) .stage1-unused.list .stage1-done.list
	@echo "Total packages:  $$(cat "$(ALL_SRPMS)" | wc -l)"
	@echo "Not found SRPMs: $$(cat .stage1-nosrpm.list | wc -l)"
	@echo
	@( \
		echo "Arch|Done|Excl|BDep|Chk|NoSI|Other|LErr|NoAtt|Total"; \
		echo "----|----|----|----|----|----|----|----|----|----"; \
		for arch in $(ARCHES); do \
			pkgs_done="$$(cat ".stage1-done.$$arch.list" | wc -l)"; \
			pkgs_excl="$$(cat ".stage1-excl.$$arch.list" | wc -l)"; \
			pkgs_brkndep="$$(cat ".stage1-brkndep.$$arch.list" | wc -l)"; \
			pkgs_chkerr="$$(cat ".stage1-chkerr.$$arch.list" | wc -l)"; \
			pkgs_nospace="$$(cat ".stage1-nospace.$$arch.list" | wc -l)"; \
			pkgs_noinode="$$(cat ".stage1-noinode.$$arch.list" | wc -l)"; \
			pkgs_logerr="$$(cat ".stage1-logerr.$$arch.list" | wc -l)"; \
			pkgs_other="$$(cat ".stage1-other.$$arch.list" | wc -l)"; \
			pkgs_noatt="$$(cat ".stage1-noatt.$$arch.list" | wc -l)"; \
			echo -n "$$arch"; \
			echo -n "|$$pkgs_done"; \
			echo -n "|$$pkgs_excl"; \
			echo -n "|$$pkgs_brkndep"; \
			echo -n "|$$pkgs_chkerr"; \
			echo -n "|$$((pkgs_nospace + pkgs_noinode))"; \
			echo -n "|$$pkgs_other"; \
			echo -n "|$$pkgs_logerr"; \
			echo -n "|$$pkgs_noatt"; \
			echo -n "|$$((pkgs_done + pkgs_excl + pkgs_brkndep + pkgs_chkerr + pkgs_nospace + pkgs_noinode + pkgs_other + pkgs_logerr + pkgs_noatt))"; \
			echo; \
		done \
	) | column -t -s '|'
	@echo
	@echo "Total unique SRPMs processed: $$(cat ".stage1-done.list" | wc -l)"
	@echo "Lists for unused packages:    $$(cat .stage1-unused.list | wc -l)"
phonytargets += stage1-status

$(STAGE1_STATUS):
	@touch $@

%.srpms: %.list
	@sed -e 's/^.*$$/&.src.rpm/' $< | sort -u >$@

$(dir $(STAGE1_LOG)):
	mkdir -p $@

$(STAGE1_LOGDIR):
	mkdir -p $@

stage1-rebuild-all: $(ALL_SRPMS) | $(dir $(STAGE1_LOG)) $(STAGE1_LOGDIR)
	$(call backup_log,$(STAGE1_LOG)$(LOG_SUF))
	-time $(HSH_SEPARATE_SOURCES) $(BUILD_COMMON_OPTS) "-s$(SKIPLIST_DIR)/" -l "$(STAGE1_LOGDIR)" $(OPTS) $(BUILDERS) $(ALL_SRPMS) 2>&1 $(if $(DRY_RUN),,| tee "$(STAGE1_LOG)$(LOG_SUF)")
	touch "$(SKIPLIST_DIR)"
	$(call import_log,$(STAGE1_LOG)$(LOG_SUF),$(STAGE1_STATUS))
phonytargets += stage1-rebuild-all

stage1-rebuild-all-dry-run:
	@$(MAKE_SELF) stage1-rebuild-all DRY_RUN=1 OPTS="--dry-run $(OPTS)"
phonytargets += stage1-rebuild-all-dry-run
stage1-rebuild-all-dry-run-run:
	@$(MAKE_SELF) stage1-rebuild-all DRY_RUN=1 OPTS="--dry-run-run $(OPTS)"
phonytargets += stage1-rebuild-all-dry-run-run

srpm_lists =
define stage1_template =
 stage1$(1): .stage1-$(2).srpms | $$(dir $$(STAGE1_LOG)) $$(STAGE1_LOGDIR)
	$$(call backup_log,$$(STAGE1_LOG)$$(LOG_SUF))
	-time $$(HSH_SEPARATE_SOURCES) $$(BUILD_COMMON_OPTS) "-s$$(SKIPLIST_DIR)/" -l "$$(STAGE1_LOGDIR)" $$(OPTS) $$(if $$(FOR_ARCHES),$$(call BUILDERS,$$(subst $$(comma), ,$$(FOR_ARCHES))),$$(BUILDERS)) .stage1-$(2).srpms 2>&1 $$(if $$(DRY_RUN),,| tee "$$(STAGE1_LOG)$$(LOG_SUF)")
	touch "$$(SKIPLIST_DIR)"
	$$(call import_log,$$(STAGE1_LOG)$$(LOG_SUF),$$(STAGE1_STATUS))
 stage1$(1)-dry-run:
	@$$(MAKE_SELF) stage1$(1) DRY_RUN=1 OPTS="--dry-run $$(OPTS)"
 stage1$(1)-dry-run-run:
	@$$(MAKE_SELF) stage1$(1) DRY_RUN=1 OPTS="--dry-run-run $$(OPTS)"
 srpm_lists += .stage1-$(2).srpms
 phonytargets += stage1$(1) stage1$(1)-dry-run stage1$(1)-dry-run-run
endef
define stage1_arch_template =
 stage1$(1)-$(3): .stage1-$(2).$(3).srpms | $$(dir $$(STAGE1_LOG)) $$(STAGE1_LOGDIR)
	$$(call backup_log,$$(STAGE1_LOG)$$(LOG_SUF))
	-time $$(HSH_SEPARATE_SOURCES) $$(BUILD_COMMON_OPTS) "-s$$(SKIPLIST_DIR)/" -l "$$(STAGE1_LOGDIR)" $$(OPTS) $$(call BUILDERS,$(3)) .stage1-$(2).$(3).srpms 2>&1 $$(if $$(DRY_RUN),,| tee "$$(STAGE1_LOG)$$(LOG_SUF)")
	touch "$$(SKIPLIST_DIR)"
	$$(call import_log,$$(STAGE1_LOG)$$(LOG_SUF),$$(STAGE1_STATUS))
 stage1$(1)-$(3)-dry-run:
	@$$(MAKE_SELF) stage1$(1)-$(3) DRY_RUN=1 OPTS="--dry-run $$(OPTS)"
 stage1$(1)-$(3)-dry-run-run:
	@$$(MAKE_SELF) stage1$(1)-$(3) DRY_RUN=1 OPTS="--dry-run-run $$(OPTS)"
 srpm_lists += .stage1-$(2).$(3).srpms
 phonytargets += stage1$(1)-$(3) stage1$(1)-$(3)-dry-run stage1$(1)-$(3)-dry-run-run
endef

retrycats1 = blderr brkndep chkerr nospace noinode logerr other

$(foreach cat,$(retrycats1),$(eval $(call stage1_template,-retry-$(cat),$(cat))))
$(eval $(call stage1_template,,noatt))
$(foreach cat,$(retrycats1),$(foreach arch,$(ARCHES),$(eval $(call stage1_arch_template,-retry-$(cat),$(cat),$(arch)))))
$(foreach arch,$(ARCHES),$(eval $(call stage1_arch_template,,noatt,$(arch))))

stage1-retry-pkgs: | $(dir $(STAGE1_LOG)) $(STAGE1_LOGDIR)
	$(call backup_log,$(STAGE1_LOG)$(LOG_SUF))
	-time $(HSH_SEPARATE_SOURCES) $(BUILD_COMMON_OPTS) "-s$(SKIPLIST_DIR)/" -l "$(STAGE1_LOGDIR)" $(OPTS) $(if $(FOR_ARCHES),$(call BUILDERS,$(subst $(comma), ,$(FOR_ARCHES))),$(BUILDERS)) $(addsuffix .src.rpm,$(PKGS)) 2>&1 $(if $(DRY_RUN),,| tee "$(STAGE1_LOG)$(LOG_SUF)")
	touch "$(SKIPLIST_DIR)"
	$(call import_log,$(STAGE1_LOG)$(LOG_SUF),$(STAGE1_STATUS))
stage1-retry-pkgs-dry-run:
	@$(MAKE_SELF) stage1-retry-pkgs DRY_RUN=1 OPTS="--dry-run $(OPTS)" PKGS="$(PKGS)" FOR_ARCHES="$(FOR_ARCHES)"
stage1-retry-pkgs-dry-run-run:
	@$(MAKE_SELF) stage1-retry-pkgs DRY_RUN=1 OPTS="--dry-run-run $(OPTS)" PKGS="$(PKGS)" FOR_ARCHES="$(FOR_ARCHES)"
phonytargets += stage1-retry-pkgs stage1-retry-pkgs-dry-run stage1-retry-pkgs-dry-run-run


### --- Stage II ---

.stage2-findlist.list: $(TAR_DIR)
	@find $(TAR_DIR) -name '*.tar' -printf '%f\n' | sed -e 's/\.tar$$//' | sort -u >$@
package_lists += .stage2-findlist.list

.stage2-unused.list: .included.list .stage2-findlist.list
	@comm -13 ".included.list" ".stage2-findlist.list" >$@
package_lists += .stage2-unused.list

clean-stage2-unused: .stage2-unused.list
	@cat $< | while read -r f; do rm -fv $(RMOPTS) "$(TAR_DIR)/$$f.tar"; done
phonytargets += clean-stage2-unused

define stage2_status_list_template
 .stage2-$(1).list: $$(STAGE2_STATUS) .included.list
	@$$(call filter_status,$$(STAGE2_STATUS),CLEAN\t$(2)) | comm -12 ".included.list" - >$$@
 package_lists += .stage2-$(1).list
endef

$(eval $(call stage2_status_list_template,statdone,TAROK))
$(eval $(call stage2_status_list_template,nochgs,NOCHGS))
$(eval $(call stage2_status_list_template,imperr,IMPERR))
$(eval $(call stage2_status_list_template,trncerr,TRNCERR))
$(eval $(call stage2_status_list_template,gearerr,GEARERR))

.stage2-done.list: .stage2-statdone.list .stage2-findlist.list
	@comm -12 ".stage2-statdone.list" ".stage2-findlist.list" >$@

.stage2-empty.list: .included.list $(SKIPLIST_DIR)
	@find "$(SKIPLIST_DIR)" -name '*.all.skiplist' -empty -printf '%f\n' | sed -e 's/\.all\.skiplist$$//' | sort -u | comm -12 ".included.list" - >$@
package_lists += .stage2-empty.list

.stage2-nonempty.list: .included.list $(SKIPLIST_DIR)
	@find "$(SKIPLIST_DIR)" -name '*.all.skiplist' ! -empty -printf '%f\n' | sed -e 's/\.all\.skiplist$$//' | sort -u | comm -12 ".included.list" - >$@
package_lists += .stage2-nonempty.list

.stage2-noatt.list: .stage1-done.list .stage2-done.list .stage2-nochgs.list .stage2-imperr.list .stage2-trncerr.list .stage2-gearerr.list .stage2-empty.list
	@cat ".stage2-done.list" ".stage2-nochgs.list" ".stage2-imperr.list" ".stage2-trncerr.list" ".stage2-gearerr.list" ".stage2-empty.list" | sort -u | comm -23 ".stage1-done.list" - >$@
package_lists += .stage2-noatt.list

statcats2 = unused done imperr trncerr gearerr nochgs empty nonempty noatt
$(foreach cat,$(statcats2),$(eval $(call print_list_template,stage2,$(cat))))

define print_logs_template =
 list-$(1)-$(2)-logs: .$(1)-$(2).list
	@cat $$< | sed -e 's:^.*$$$$:$$($(patsubst %:$(1),%,$(filter %:$(1),STAGE1_LOGDIR:stage1 STAGE2_LOGDIR:stage2 STAGE3_LOGDIR:stage3 STAGE4_LOGDIR:stage4 STAGE5_V_LOGDIR:stage5v STAGE5_O_LOGDIR:stage5o)))/&.log:' | sort -u
 phonytargets += list-$(1)-$(2)-logs
endef
$(foreach cat,$(statcats2),$(eval $(call print_logs_template,stage2,$(cat))))

stage2-status: $(foreach cat,$(statcats2),.stage2-$(cat).list)
	@echo "Non-empty united skiplists: $$(cat ".stage2-nonempty.list" | wc -l)"
	@echo "Empty united skiplists:     $$(cat ".stage2-empty.list" | wc -l)"
	@echo "Total:                      $$(sort -u ".stage2-empty.list" ".stage2-nonempty.list" | wc -l)"
	@echo
	@( \
		echo "Done|NoChgs|Imp.err|Trnc.err|Gr.err|NoAtt|Total"; \
		echo "----|----|----|----|----|----|----"; \
		pkgs_done="$$(cat ".stage2-done.list" | wc -l)"; \
		pkgs_nochgs="$$(cat ".stage2-nochgs.list" | wc -l)"; \
		pkgs_imperr="$$(cat ".stage2-imperr.list" | wc -l)"; \
		pkgs_trncerr="$$(cat ".stage2-trncerr.list" | wc -l)"; \
		pkgs_gearerr="$$(cat ".stage2-gearerr.list" | wc -l)"; \
		pkgs_noatt="$$(cat ".stage2-noatt.list" | wc -l)"; \
		echo -n "$$pkgs_done"; \
		echo -n "|$$pkgs_nochgs"; \
		echo -n "|$$pkgs_imperr"; \
		echo -n "|$$pkgs_trncerr"; \
		echo -n "|$$pkgs_gearerr"; \
		echo -n "|$$pkgs_noatt"; \
		echo -n "|$$((pkgs_done + pkgs_nochgs + pkgs_imperr + pkgs_trncerr + pkgs_gearerr + pkgs_noatt))"; \
		echo; \
	) | column -t -s '|'
	@echo "Unused cleaned-up packages:  $$(cat ".stage2-unused.list" | wc -l)"
phonytargets += stage2-status

$(STAGE2_STATUS):
	@touch $@

ifneq ($(dir $(STAGE2_LOG)),$(dir $(STAGE1_LOG)))
$(dir $(STAGE2_LOG)):
	mkdir -p $@
endif

$(STAGE2_LOGDIR):
	mkdir -p $@

stage2-rebuild-all: | $(SKIPLIST_DIR) $(dir $(STAGE2_LOG)) $(STAGE2_LOGDIR)
	$(call backup_log,$(STAGE2_LOG)$(LOG_SUF))
	-time $(SRPM_CLEANUP) $(CLEAN_COMMON_OPTS) $(if $(EXCEPTIONS),-x "$(EXCEPTIONS)") -t -T "$(TAR_DIR)" -l "$(STAGE2_LOGDIR)" $(OPTS) $(CLEANERS) $(SKIPLIST_DIR) 2>&1 $(if $(DRY_RUN),,| tee "$(STAGE2_LOG)$(LOG_SUF)")
	touch "$(TAR_DIR)"
	$(call import_log_noarch,$(STAGE2_LOG)$(LOG_SUF),$(STAGE2_STATUS))
phonytargets += stage2-rebuild-all

stage2-rebuild-all-dry-run:
	@$(MAKE_SELF) stage2-rebuild-all DRY_RUN=1 OPTS="--dry-run $(OPTS)"
phonytargets += stage2-rebuild-all-dry-run
stage2-rebuild-all-dry-run-run:
	@$(MAKE_SELF) stage2-rebuild-all DRY_RUN=1 OPTS="--dry-run-run $(OPTS)"
phonytargets += stage2-rebuild-all-dry-run-run

%.skiplist.all: %.list
	@sed -e 's:^.*$$:$(foreach arch,$(ARCHES),\n$(SKIPLIST_DIR)/&.$(arch).skiplist):' -e 's/ \n/\n/g' -e 's/^\n//' $< | sort -u >$@

%.skiplist.some: %.skiplist.all
	@find "$(SKIPLIST_DIR)" -name '*.skiplist' -a -not -name '*.all.skiplist' | grep -Fx -f $< | sort -u >$@

skiplists =

define stage2_template =
 stage2$(1): .stage2-$(2).skiplist.some | $$(dir $$(STAGE2_LOG)) $$(STAGE2_LOGDIR)
	$$(call backup_log,$$(STAGE2_LOG)$$(LOG_SUF))
	-time $$(SRPM_CLEANUP) $$(CLEAN_COMMON_OPTS) $$(if $$(EXCEPTIONS),-x "$$(EXCEPTIONS)") -t -T "$$(TAR_DIR)" -l "$$(STAGE2_LOGDIR)" $$(OPTS) $$(CLEANERS) $$< 2>&1 $$(if $$(DRY_RUN),,| tee "$$(STAGE2_LOG)$$(LOG_SUF)")
	touch "$$(TAR_DIR)"
	$$(call import_log_noarch,$$(STAGE2_LOG)$$(LOG_SUF),$$(STAGE2_STATUS))
 stage2$(1)-dry-run:
	@$$(MAKE_SELF) stage2$(1) DRY_RUN=1 OPTS="--dry-run $$(OPTS)"
 stage2$(1)-dry-run-run:
	@$$(MAKE_SELF) stage2$(1) DRY_RUN=1 OPTS="--dry-run-run $$(OPTS)"
 skiplists += .stage2-$(2).skiplist.all .stage2-$(2).skiplist.some
 phonytargets += stage2$(1) stage2$(1)-dry-run stage2$(1)-dry-run-run
endef

retrycats2 = imperr trncerr gearerr
$(foreach cat,$(retrycats2),$(eval $(call stage2_template,-retry-$(cat),$(cat))))
$(eval $(call stage2_template,,noatt))

stage2-retry-pkgs: | $(dir $(STAGE2_LOG)) $(STAGE2_LOGDIR)
	$(call backup_log,$(STAGE2_LOG)$(LOG_SUF))
	-time $(SRPM_CLEANUP) $(CLEAN_COMMON_OPTS) $(if $(EXCEPTIONS),-x "$(EXCEPTIONS)") -t -T "$(TAR_DIR)" -l "$(STAGE2_LOGDIR)" -s "$(SKIPLIST_DIR)" $(OPTS) $(CLEANERS) $(PKGS) 2>&1 $(if $(DRY_RUN),,| tee "$(STAGE2_LOG)$(LOG_SUF)")
	touch "$(TAR_DIR)"
	$(call import_log_noarch,$(STAGE2_LOG)$(LOG_SUF),$(STAGE2_STATUS))
stage2-retry-pkgs-dry-run:
	@$(MAKE_SELF) stage2-retry-pkgs DRY_RUN=1 OPTS="--dry-run $(OPTS)" PKGS="$(PKGS)"
stage2-retry-pkgs-dry-run-run:
	@$(MAKE_SELF) stage2-retry-pkgs DRY_RUN=1 OPTS="--dry-run-run $(OPTS)" PKGS="$(PKGS)"
phonytargets += stage2-retry-pkgs stage2-retry-pkgs-dry-run stage2-retry-pkgs-dry-run-run


### --- Stage III ---

.stage3-findlist.list: $(SRPM_DIR)
	@find $(SRPM_DIR) -name '*.src.rpm' -printf '%f\n' | sed -e 's/\.src\.rpm$$//' | sort -u >$@
package_lists += .stage3-findlist.list

.stage3-unused.list: .included.list .stage3-findlist.list
	@comm -13 ".included.list" ".stage3-findlist.list" >$@
package_lists += .stage3-unused.list

clean-stage3-unused: .stage3-unused.list
	@cat $< | while read -r f; do rm -fv $(RMOPTS) "$(SRPM_DIR)/$$f.src.rpm"; done
phonytargets += clean-stage3-unused

define stage3_status_list_template
 .stage3-$(1).list: $$(STAGE3_STATUS) .included.list
	@$$(call filter_status,$$(STAGE3_STATUS),SRPM\t$(2))  | comm -12 ".included.list" - >$$@
 package_lists += .stage3-$(1).list
endef

$(eval $(call stage3_status_list_template,statdone,BLDOK\t\tCPOK))
$(eval $(call stage3_status_list_template,enverr,ENVERR))
$(eval $(call stage3_status_list_template,brkndep,ENVERR\tBRKNDEP))
$(eval $(call stage3_status_list_template,blderr,BLDERR))
$(eval $(call stage3_status_list_template,nospace,BLDERR\tNOSPACE))
$(eval $(call stage3_status_list_template,noinode,BLDERR\tNOINODE))
$(eval $(call stage3_status_list_template,nocopy,BLDOK\t\tCPERR))

.stage3-done.list: .stage3-statdone.list .stage3-findlist.list
	@comm -12 ".stage3-statdone.list" ".stage3-findlist.list" >$@
package_lists += .stage3-done.list

.stage3-eother.list: .stage3-enverr.list .stage3-brkndep.list
	@cat ".stage3-brkndep.list" | sort -u | comm -23 ".stage3-enverr.list" - >$@
.stage3-bother.list: .stage3-blderr.list .stage3-nospace.list .stage3-noinode.list
	@cat ".stage3-nospace.list" ".stage3-noinode.list" | sort -u | comm -23 ".stage3-blderr.list" - >$@
.stage3-other.list: .stage3-eother.list .stage3-bother.list
	@sort -u $^ >$@
package_lists += .stage3-eother.list .stage3-bother.list .stage3-other.list

.stage3-errs.list: .stage3-enverr.list .stage3-blderr.list .stage3-nocopy.list
	@sort -u $^ >$@
package_lists += .stage3-errs.list

.stage3-noatt.list: .stage2-done.list .stage3-done.list .stage3-enverr.list .stage3-blderr.list .stage3-nocopy.list
	@cat ".stage3-done.list" ".stage3-enverr.list" ".stage3-blderr.list" ".stage3-nocopy.list" | sort -u | comm -23 ".stage2-done.list" - >$@
package_lists += .stage3-noatt.list

statcats3 = unused done enverr brkndep blderr nospace noinode nocopy other errs noatt

$(foreach cat,$(statcats3),$(eval $(call print_list_template,stage3,$(cat))))
$(foreach cat,$(statcats3),$(eval $(call print_logs_template,stage3,$(cat))))

stage3-status: $(foreach cat,$(statcats3),.stage3-$(cat).list) .stage2-done.list
	@echo "Total cleaned-up TAR packages: $$(cat ".stage2-done.list" | wc -l)"
	@echo
	@( \
		echo "Done|BDep|NoSI|Other|NoCopy|NoAtt|Total"; \
		echo "----|----|----|----|----|----|----"; \
		pkgs_done="$$(cat ".stage3-done.list" | wc -l)"; \
		pkgs_brkndep="$$(cat ".stage3-brkndep.list" | wc -l)"; \
		pkgs_nospace="$$(cat ".stage3-nospace.list" | wc -l)"; \
		pkgs_noinode="$$(cat ".stage3-noinode.list" | wc -l)"; \
		pkgs_other="$$(cat ".stage3-other.list" | wc -l)"; \
		pkgs_nocopy="$$(cat ".stage3-nocopy.list" | wc -l)"; \
		pkgs_noatt="$$(cat ".stage3-noatt.list" | wc -l)"; \
		echo -n "$$pkgs_done"; \
		echo -n "|$$pkgs_brkndep"; \
		echo -n "|$$((pkgs_nospace + pkgs_noinode))"; \
		echo -n "|$$pkgs_other"; \
		echo -n "|$$pkgs_nocopy"; \
		echo -n "|$$pkgs_noatt"; \
		echo -n "|$$((pkgs_done + pkgs_brkndep + pkgs_nospace + pkgs_noinode + pkgs_other + pkgs_nocopy + pkgs_noatt))"; \
		echo; \
	) | column -t -s '|'
	@echo
	@echo "Total failed packages:    $$(cat ".stage3-errs.list" | wc -l)"
	@echo "Unused cleaned-up SRPMs:  $$(cat ".stage3-unused.list" | wc -l)"
phonytargets += stage3-status

$(STAGE3_STATUS):
	@touch $@

ifneq ($(dir $(STAGE3_LOG)),$(dir $(STAGE2_LOG)))
$(dir $(STAGE3_LOG)):
	mkdir -p $@
endif

$(STAGE3_LOGDIR):
	mkdir -p $@

stage3-rebuild-all: | $(SKIPLIST_DIR) $(TAR_DIR) $(dir $(STAGE3_LOG)) $(STAGE3_LOGDIR)
	$(call backup_log,$(STAGE3_LOG)$(LOG_SUF))
	-time $(SRPM_CLEANUP) $(CLEAN_COMMON_OPTS) $(if $(EXCEPTIONS),-x "$(EXCEPTIONS)") -T "$(TAR_DIR)" --with-tars -O "$(SRPM_DIR)" -l "$(STAGE3_LOGDIR)" $(OPTS) $(SRPM_BUILDERS) $(SKIPLIST_DIR) 2>&1 $(if $(DRY_RUN),,| tee "$(STAGE3_LOG)$(LOG_SUF)")
	touch "$(SRPM_DIR)"
	$(call import_log_noarch,$(STAGE3_LOG)$(LOG_SUF),$(STAGE3_STATUS))
phonytargets += stage3-rebuild-all

stage3-rebuild-all-dry-run:
	@$(MAKE_SELF) stage3-rebuild-all DRY_RUN=1 OPTS="--dry-run $(OPTS)"
phonytargets += stage3-rebuild-all-dry-run
stage3-rebuild-all-dry-run-run:
	@$(MAKE_SELF) stage3-rebuild-all DRY_RUN=1 OPTS="--dry-run-run $(OPTS)"
phonytargets += stage3-rebuild-all-dry-run-run

define stage3_template =
 stage3$(1): .stage3-$(2).skiplist.some | $$(dir $$(STAGE3_LOG)) $$(STAGE3_LOGDIR)
	$$(call backup_log,$$(STAGE3_LOG)$(LOG_SUF))
	-time $$(SRPM_CLEANUP) $$(CLEAN_COMMON_OPTS) $$(if $$(EXCEPTIONS),-x "$$(EXCEPTIONS)") -T "$$(TAR_DIR)" --with-tars -O "$$(SRPM_DIR)" -l "$$(STAGE3_LOGDIR)" $$(OPTS) $$(SRPM_BUILDERS) $$< 2>&1 $$(if $$(DRY_RUN),,| tee "$$(STAGE3_LOG)$$(LOG_SUF)")
	touch "$$(SRPM_DIR)"
	$$(call import_log_noarch,$$(STAGE3_LOG)$$(LOG_SUF),$$(STAGE3_STATUS))
 stage3$(1)-dry-run:
	@$$(MAKE_SELF) stage3$(1) DRY_RUN=1 OPTS="--dry-run $$(OPTS)"
 stage3$(1)-dry-run-run:
	@$$(MAKE_SELF) stage3$(1) DRY_RUN=1 OPTS="--dry-run-run $$(OPTS)"
 skiplists += .stage3-$(2).skiplist.all .stage3-$(2).skiplist.some
 phonytargets += stage3$(1) stage3$(1)-dry-run stage3$(1)-dry-run-run
endef

retrycats3 = blderr brkndep nospace noinode nocopy errs other
$(foreach cat,$(retrycats3),$(eval $(call stage3_template,-retry-$(cat),$(cat))))
$(eval $(call stage3_template,,noatt))

stage3-retry-pkgs: | $(dir $(STAGE3_LOG)) $(STAGE3_LOGDIR)
	$(call backup_log,$(STAGE3_LOG)$(LOG_SUF))
	-time $(SRPM_CLEANUP) $(CLEAN_COMMON_OPTS) $(if $(EXCEPTIONS),-x "$(EXCEPTIONS)") -T "$(TAR_DIR)" --with-tars -O "$(SRPM_DIR)" -l "$(STAGE3_LOGDIR)" -s "$(SKIPLIST_DIR)" $(OPTS) $(SRPM_BUILDERS) $(PKGS) 2>&1 $(if $(DRY_RUN),,| tee "$(STAGE3_LOG)$(LOG_SUF)")
	touch "$(SRPM_DIR)"
	$(call import_log_noarch,$(STAGE3_LOG)$(LOG_SUF),$(STAGE3_STATUS))
stage3-retry-pkgs-dry-run:
	@$(MAKE_SELF) stage3-retry-pkgs DRY_RUN=1 OPTS="--dry-run $(OPTS)" PKGS="$(PKGS)"
stage3-retry-pkgs-dry-run-run:
	@$(MAKE_SELF) stage3-retry-pkgs DRY_RUN=1 OPTS="--dry-run-run $(OPTS)" PKGS="$(PKGS)"
phonytargets += stage3-retry-pkgs stage3-retry-pkgs-dry-run stage3-retry-pkgs-dry-run-run


### --- Stage IV ---

define stage4_unused_arch_list_template =
 .stage4-findlist.$(1).list: $$(RPM_DIR)
	@find $$(RPM_DIR) -maxdepth 1 -type d -name "*.$(1)" -printf '%f\n' | sed -e 's/\.[^.]\+$$$$//' | sort -u >$$@
 package_lists += .stage4-findlist.$(1).list

 .stage4-unused.$(1).list: .included.$(1).list .stage4-findlist.$(1).list
	@comm -13 ".included.$(1).list" ".stage4-findlist.$(1).list" >$$@
 package_lists += .stage4-unused.$(1).list

 clean-stage4-$(1)-unused: .stage4-unused.$(1).list
	@cat $$< | while read -r d; do rm -rfv $$(RMOPTS) "$$(RPM_DIR)/$$$$d.$(1)"; done
 phonytargets += clean-stage4-$(1)-unused
endef
$(foreach arch,$(ARCHES),$(eval $(call stage4_unused_arch_list_template,$(arch))))

define stage4_arch_status_list_template =
 .stage4-$(1).$(2).list: $$(STAGE4_STATUS) .included.$(2).list
	@$$(call filter_arch_status,$$(STAGE4_STATUS),$(2),REBUILD\t$(3)) | comm -12 ".included.$(2).list" - >$$@
 package_lists += .stage4-$(1).$(2).list
endef
define stage4_status_list_template =
 $$(foreach arch,$$(ARCHES),$$(eval $$(call stage4_arch_status_list_template,$(1),$$(arch),$(2))))
endef

$(eval $(call stage4_status_list_template,statdone,\(LOGOK\tVFYOK\tCPOK\|ALRDY\)))
$(eval $(call stage4_status_list_template,blderr,BLDERR))
$(eval $(call stage4_status_list_template,brkndep,BLDERR\tBRKNDEP))
$(eval $(call stage4_status_list_template,chkerr,BLDERR\tCHKERR))
$(eval $(call stage4_status_list_template,nospace,BLDERR\tNOSPACE))
$(eval $(call stage4_status_list_template,noinode,BLDERR\tNOINODE))
$(eval $(call stage4_status_list_template,logerr,LOGERR))
$(eval $(call stage4_status_list_template,vfyerr,LOGERR\tVFYERR))
$(eval $(call stage4_status_list_template,nocopy,LOGOK\t\t\(EXPERR\|CPERR\)))
$(eval $(call stage4_status_list_template,unknown,\t\t))

define stage4_arch_lists_template =
 .stage4-bother.$(1).list: .stage4-blderr.$(1).list .stage4-brkndep.$(1).list .stage4-chkerr.$(1).list .stage4-nospace.$(1).list .stage4-noinode.$(1).list
	@cat ".stage4-brkndep.$(1).list" ".stage4-chkerr.$(1).list" ".stage4-nospace.$(1).list" ".stage4-noinode.$(1).list" | sort -u | comm -23 ".stage4-blderr.$(1).list" - >$$@
 .stage4-lother.$(1).list: .stage4-logerr.$(1).list .stage4-vfyerr.$(1).list
	@comm -23 ".stage4-logerr.$(1).list" ".stage4-vfyerr.$(1).list" >$$@
 .stage4-other.$(1).list: .stage4-bother.$(1).list .stage4-lother.$(1).list .stage4-unknown.$(1).list
	@sort -u $$^ >$$@
 .stage1-3-check.$(1).list: .stage1-done.$(1).list .stage2-done.list .stage3-done.list .stage2-empty.list .stage2-nochgs.list
	@comm -23 ".stage1-done.$(1).list" ".stage2-empty.list" | comm -13 ".stage2-nochgs.list" - | comm -23 ".stage2-done.list" - | comm -23 ".stage3-done.list" - >$$@
 .stage4-noatt.$(1).list: .stage1-3-check.$(1).list .stage4-blderr.$(1).list .stage4-logerr.$(1).list .stage4-nocopy.$(1).list .stage4-done.$(1).list .stage4-unknown.$(1).list
	@cat ".stage4-done.$(1).list" ".stage4-blderr.$(1).list" ".stage4-logerr.$(1).list" ".stage4-nocopy.$(1).list" ".stage4-unknown.$(1).list" | sort -u | comm -23 ".stage1-3-check.$(1).list" - >$$@
 .stage4-done.$(1).list: .stage4-statdone.$(1).list .stage4-findlist.$(1).list
	@comm -12 ".stage4-statdone.$(1).list" ".stage4-findlist.$(1).list" >$$@
 package_lists += .stage4-bother.$(1).list .stage4-lother.$(1).list .stage4-other.$(1).list .stage1-3-check.$(1).list .stage4-noatt.$(1).list .stage4-done.$(1).list
endef
$(foreach arch,$(ARCHES),$(eval $(call stage4_arch_lists_template,$(arch))))

statcats4 = unused done blderr brkndep chkerr nospace noinode logerr vfyerr nocopy other noatt

$(foreach cat,$(statcats4),$(eval $(call print_list_template,stage4,$(cat))))
$(foreach cat,$(statcats4),$(foreach arch,$(ARCHES),$(eval $(call print_arch_list_template,stage4,$(cat),$(arch)))))
$(foreach cat,$(statcats4),$(foreach arch,$(ARCHES),$(eval $(call print_arch_logs_template,stage4,$(cat),$(arch)))))
$(foreach cat,$(statcats4),$(eval $(call united_list_template,.stage4-$(cat))))

stage4-status: .stage3-done.list $(foreach cat,$(statcats4),$(foreach arch,$(ARCHES),.stage4-$(cat).$(arch).list)) $(foreach arch,$(ARCHES),.stage1-3-check.$(arch).list) .stage4-unused.list .stage4-done.list
	@echo "Total cleaned-up SRPM packages:  $$(cat ".stage3-done.list" | wc -l)"
	@echo
	@( \
		echo "Arch|Done|BDep|Chk|NoSI|Other|VErr|NoAtt|Total|I-III"; \
		echo "----|----|----|----|----|----|----|----|----|-----"; \
		for arch in $(ARCHES); do \
			pkgs_done="$$(cat ".stage4-done.$$arch.list" | wc -l)"; \
			pkgs_brkndep="$$(cat ".stage4-brkndep.$$arch.list" | wc -l)"; \
			pkgs_chkerr="$$(cat ".stage4-chkerr.$$arch.list" | wc -l)"; \
			pkgs_nospace="$$(cat ".stage4-nospace.$$arch.list" | wc -l)"; \
			pkgs_noinode="$$(cat ".stage4-noinode.$$arch.list" | wc -l)"; \
			pkgs_other="$$(cat ".stage4-other.$$arch.list" | wc -l)"; \
			pkgs_vfyerr="$$(cat ".stage4-vfyerr.$$arch.list" | wc -l)"; \
			pkgs_noatt="$$(cat ".stage4-noatt.$$arch.list" | wc -l)"; \
			pkgs_stg13check="$$(cat ".stage1-3-check.$$arch.list" | wc -l)"; \
			echo -n "$$arch"; \
			echo -n "|$$pkgs_done"; \
			echo -n "|$$pkgs_brkndep"; \
			echo -n "|$$pkgs_chkerr"; \
			echo -n "|$$((pkgs_nospace + pkgs_noinode))"; \
			echo -n "|$$pkgs_other"; \
			echo -n "|$$pkgs_vfyerr"; \
			echo -n "|$$pkgs_noatt"; \
			echo -n "|$$((pkgs_done + pkgs_brkndep + pkgs_chkerr + pkgs_nospace + pkgs_noinode + pkgs_other + pkgs_vfyerr + pkgs_noatt))"; \
			echo -n "|$$pkgs_stg13check"; \
			echo; \
		done \
	) | column -t -s '|'
	@echo
	@echo "Total unique cleaned-up SRPMs verified:  $$(cat ".stage4-done.list" | wc -l)"
	@echo "Unused verified SRPMs:  $$(cat ".stage4-unused.list" | wc -l)"
phonytargets += stage4-status

$(STAGE4_STATUS):
	@touch $@

ifneq ($(dir $(STAGE4_LOG)),$(dir $(STAGE3_LOG)))
$(dir $(STAGE4_LOG)):
	mkdir -p $@
endif

$(STAGE4_LOGDIR):
	mkdir -p $@

stage4-rebuild-all: $(if $(DISTTAG_OPTS),$(DISTTAG_INDEX)) | $(SRPM_DIR) $(dir $(STAGE4_LOG)) $(STAGE4_LOGDIR)
	$(call backup_log,$(STAGE4_LOG)$(LOG_SUF))
	-time $(HSH_SEPARATE_SOURCES) $(BUILD_COMMON_OPTS) -y $(if $(EXCEPTIONS),-x "$(EXCEPTIONS)") $(if $(TOOLERANCE),-t "$(TOOLERANCE)") "-s$(SKIPLIST_DIR)/" -O "$(RPM_DIR)" -l "$(STAGE4_LOGDIR)" $(DISTTAG_OPTS) $(OPTS) -B "$(SRPM_DIR)" $(STAGE4_BUILDERS) $(SRPM_DIR) 2>&1 $(if $(DRY_RUN),,| tee "$(STAGE4_LOG)$(LOG_SUF)")
	touch "$(RPM_DIR)"
	$(call import_log,$(STAGE4_LOG)$(LOG_SUF),$(STAGE4_STATUS))
phonytargets += stage4-rebuild-all

stage4-rebuild-all-dry-run:
	@$(MAKE_SELF) stage4-rebuild-all DRY_RUN=1 OPTS="--dry-run $(OPTS)"
phonytargets += stage4-rebuild-all-dry-run
stage4-rebuild-all-dry-run-run:
	@$(MAKE_SELF) stage4-rebuild-all DRY_RUN=1 OPTS="--dry-run-run $(OPTS)"
phonytargets += stage4-rebuild-all-dry-run-run

define stage4_template =
 stage4$(1): $$(patsubst %,.stage4-$(2).%.srpms,$$(ARCHES)) $$(if $$(DISTTAG_OPTS),$$(DISTTAG_INDEX)) | $$(SRPM_DIR) $$(dir $$(STAGE4_LOG)) $$(STAGE4_LOGDIR)
	$$(call backup_log,$$(STAGE4_LOG)$$(LOG_SUF))
	-time $$(HSH_SEPARATE_SOURCES) $$(BUILD_COMMON_OPTS) -y $$(if $$(EXCEPTIONS),-x "$$(EXCEPTIONS)") $$(if $$(TOOLERANCE),-t "$$(TOOLERANCE)") "-s$$(SKIPLIST_DIR)/" -O "$$(RPM_DIR)" -l "$$(STAGE4_LOGDIR)" $$(DISTTAG_OPTS) $$(OPTS) -B "$$(SRPM_DIR)"  $$(if $$(FOR_ARCHES),$$(call STAGE4_BUILDERS,$$(subst $$(comma), ,$$(FOR_ARCHES))),$$(STAGE4_BUILDERS)) $$(patsubst %,.stage4-$(2).%.srpms,$$(ARCHES)) 2>&1 $$(if $$(DRY_RUN),,| tee "$$(STAGE4_LOG)$$(LOG_SUF)")
	touch "$$(RPM_DIR)"
	$$(call import_log,$$(STAGE4_LOG)$$(LOG_SUF),$$(STAGE4_STATUS))
 stage4$(1)-dry-run:
	@$$(MAKE_SELF) stage4$(1) DRY_RUN=1 OPTS="--dry-run $$(OPTS)"
 stage4$(1)-dry-run-run:
	@$$(MAKE_SELF) stage4$(1) DRY_RUN=1 OPTS="--dry-run-run $$(OPTS)"
 srpm_lists += $$(patsubst %,.stage4-$(2).%.srpms,$$(ARCHES))
 phonytargets += stage4$(1) stage4$(1)-dry-run stage4$(1)-dry-run-run
endef
define stage4_arch_template =
 stage4$(1)-$(3): .stage4-$(2).$(3).srpms $$(if $$(DISTTAG_OPTS),$$(DISTTAG_INDEX)) | $$(SRPM_DIR) $$(dir $$(STAGE4_LOG)) $$(STAGE4_LOGDIR)
	$$(call backup_log,$$(STAGE4_LOG)$$(LOG_SUF))
	-time $$(HSH_SEPARATE_SOURCES) $$(BUILD_COMMON_OPTS) -y $$(if $$(EXCEPTIONS),-x "$$(EXCEPTIONS)") $$(if $$(TOOLERANCE),-t "$$(TOOLERANCE)") "-s$$(SKIPLIST_DIR)/" -O "$$(RPM_DIR)" -l "$$(STAGE4_LOGDIR)" $$(DISTTAG_OPTS) $$(OPTS) -B "$$(SRPM_DIR)" $$(call STAGE4_BUILDERS,$(3)) .stage4-$(2).$(3).srpms 2>&1 $$(if $$(DRY_RUN),,| tee "$$(STAGE4_LOG)$$(LOG_SUF)")
	touch "$$(RPM_DIR)"
	$$(call import_log,$$(STAGE4_LOG)$$(LOG_SUF),$$(STAGE4_STATUS))
 stage4$(1)-$(3)-dry-run:
	@$$(MAKE_SELF) stage4$(1)-$(3) DRY_RUN=1 OPTS="--dry-run $$(OPTS)"
 stage4$(1)-$(3)-dry-run-run:
	@$$(MAKE_SELF) stage4$(1)-$(3) DRY_RUN=1 OPTS="--dry-run-run $$(OPTS)"
 srpm_lists += .stage4-$(2).$(3).srpms
 phonytargets += stage4$(1)-$(3) stage4$(1)-$(3)-dry-run stage4$(1)-$(3)-dry-run-run
endef

retrycats4 = blderr brkndep chkerr nospace noinode vfyerr other

$(foreach cat,$(retrycats4),$(eval $(call stage4_template,-retry-$(cat),$(cat))))
$(eval $(call stage4_template,,noatt))
$(foreach cat,$(retrycats4),$(foreach arch,$(ARCHES),$(eval $(call stage4_arch_template,-retry-$(cat),$(cat),$(arch)))))
$(foreach arch,$(ARCHES),$(eval $(call stage4_arch_template,,noatt,$(arch))))

stage4-retry-pkgs: $(if $(DISTTAG_OPTS),$(DISTTAG_INDEX)) | $(SRPM_DIR) $(dir $(STAGE4_LOG)) $(STAGE4_LOGDIR)
	$(call backup_log,$(STAGE4_LOG)$(LOG_SUF))
	-time $(HSH_SEPARATE_SOURCES) $(BUILD_COMMON_OPTS) -y $(if $(EXCEPTIONS),-x "$(EXCEPTIONS)") $(if $(TOOLERANCE),-t "$(TOOLERANCE)") "-s$(SKIPLIST_DIR)/" -O "$(RPM_DIR)" -l "$(STAGE4_LOGDIR)" $(DISTTAG_OPTS) $(OPTS) -B "$(SRPM_DIR)" $(if $(FOR_ARCHES),$(call STAGE4_BUILDERS,$(subst $(comma), ,$(FOR_ARCHES))),$(STAGE4_BUILDERS)) $(addsuffix .src.rpm,$(PKGS)) 2>&1 $(if $(DRY_RUN),,| tee "$(STAGE4_LOG)$(LOG_SUF)")
	touch "$(RPM_DIR)"
	$(call import_log,$(STAGE4_LOG)$(LOG_SUF),$(STAGE4_STATUS))
stage4-retry-pkgs-dry-run:
	@$(MAKE_SELF) stage4-retry-pkgs DRY_RUN=1 OPTS="--dry-run $(OPTS)" PKGS="$(PKGS)" FOR_ARCHES="$(FOR_ARCHES)"
stage4-retry-pkgs-dry-run-run:
	@$(MAKE_SELF) stage4-retry-pkgs DRY_RUN=1 OPTS="--dry-run-run $(OPTS)" PKGS="$(PKGS)" FOR_ARCHES="$(FOR_ARCHES)"
phonytargets += stage4-retry-pkgs stage4-retry-pkgs-dry-run stage4-retry-pkgs-dry-run-run


### Preparing the Stage V

$(eval $(call united_list_template,.stage4-done-included))

list-stage4-done-included: .stage4-done-included.list
	@cat $<
phonytargets += list-stage4-done-included

$(eval $(call united_list_template,.stage4-not-done-included))

list-stage4-not-done-included: .stage4-not-done-included.list
	@cat $<
phonytargets += list-stage4-not-done-included

list-stage4-done-all: .stage4-done-all.list
	@cat $<
package_lists += .stage4-done-all.list
phonytargets += list-stage4-done-all

list-stage4-done-all-files: .stage4-done-all.srpms
	@cat $<
srpm_lists += .stage4-done-all.srpms
phonytargets += list-stage4-done-all-files

.stage4-done-included-all.list: .stage4-done-all.list .included-all.list .stage4-done-exclude.list
	@comm -12 .included-all.list .stage4-done-all.list | comm -13 .stage4-done-exclude.list - >$@
package_lists += .stage4-done-included-all.list .included-all.list

list-stage4-done-included-all: .stage4-done-included-all.list
	@cat $<
phonytargets += list-stage4-done-included-all

list-stage4-done-included-all-files: .stage4-done-included-all.srpms
	@cat $<
srpm_lists += .stage4-done-included-all.srpms
phonytargets += list-stage4-done-included-all-files

export_srpm_lists =
stage4-done-included.all.srpms: .stage4-done-included-all.srpms
	cat $< >$@
export_srpm_lists += stage4-done-included.all.srpms

define stage4_done_exclusive_template
 .stage4-done-exsv.$(1).list: .stage4-done.$(1).list .stage4-done-all.list $(foreach arch,$(ARCHES),.stage4-done.$(arch).list) | $(REPO_DIR)
	@if [ -n "$$(REPO_DIR)" ] && [ -d "$$(REPO_DIR)" ]; then \
		comm -23 .stage4-done.$(1).list .stage4-done-all.list | \
			while read -r nvr; do \
				earches=; \
				for arch in $$(ARCHES); do \
					if grep -qwF "$$$$nvr" .stage4-done.$$$$arch.list; \
					then \
						[ -z "$$$$earches" ] || earches="$$$$earches "; \
						earches="$$$$earches$$$$arch"; \
					fi; \
				done; \
				if [ "$$$$earches" = "$$(strip $$(ARCHES))" ]; then \
					echo "BUG! $$$$nvr is verified on all arches!" >&2; \
					exit 1; \
				fi; \
				rarches=; \
				for arch in $$(ARCHES); do \
					if [ -e "$$(REPO_DIR)/$$$$arch/SRPMS.classic/$$$$nvr.src.rpm" ]; \
					then \
						[ -z "$$$$rarches" ] || rarches="$$$$rarches "; \
						rarches="$$$$rarches$$$$arch"; \
					fi; \
				done; \
				[ "$$$$earches" = "$$$$rarches" ] || continue; \
				echo "$$$$nvr"; \
			done; \
	else \
		echo "Please specify the correct REPO_DIR." >&2; \
		exit 1; \
	fi >$$@
 package_lists += .stage4-done-exsv.$(1).list
 list-stage4-done-exsv-$(1): .stage4-done-exsv.$(1).list
	@cat $$<
 phonytargets += list-stage4-done-exsv-$(1)
 list-stage4-done-exsv-$(1)-files: .stage4-done-exsv.$(1).srpms
	@cat $$<
 phonytargets += list-stage4-done-exsv-$(1)-files
 srpm_lists += .stage4-done-exsv.$(1).srpms
 .stage4-done-xincluded.$(1).list: .stage4-done.$(1).list .included.$(1).list
	@comm -12 .stage4-done.$(1).list .included.$(1).list >$$@
 package_lists += .stage4-done-xincluded.$(1).list
 list-stage4-done-xincluded-$(1): .stage4-done-xincluded.$(1).list
	@cat $$<
 phonytargets += list-stage4-done-xincluded-$(1)
 list-stage4-done-xincluded-$(1)-files: .stage4-done-xincluded.$(1).srpms
	@cat $$<
 phonytargets += list-stage4-done-xincluded-$(1)-files
 srpm_lists += .stage4-done-xincluded.$(1).srpms
 stage4-done-xincluded.$(1).srpms: .stage4-done-xincluded.$(1).srpms
	cat $$< >$$@
 export_srpm_lists += stage4-done-xincluded.$(1).srpms
endef
$(foreach arch,$(ARCHES),$(eval $(call stage4_done_exclusive_template,$(arch))))

scripts =
exclude.regexp: $(STAGE5_V_EXCLUDE_LIST)
	@sed -e 's/^.*$$/^&-[^-]\\+-[^-]\\+$$/' $< >$@
scripts += exclude.regexp

.stage4-done-exclude.list: $(if $(STAGE5_V_EXCLUDE_LIST),$(if $(wildcard $(STAGE5_V_EXCLUDE_LIST)),exclude.regexp),_always_) $(ALL_SRPMS)
	@$(if $(STAGE5_V_EXCLUDE_LIST),$(if $(wildcard $(STAGE5_V_EXCLUDE_LIST)),grep -f "exclude.regexp" $(ALL_SRPMS) | sed -e 's/\.src\.rpm//',echo -n),echo -n) >$@
package_lists += .stage4-done-exclude.list

define stage4_arch_inc_template
 .stage4-done-all-plus-exsv.$(1).list: .stage4-done-all.list .stage4-done.$(1).list .stage4-done-exclude.list
	@cat $$^ | sort -u | comm -13 .stage4-done-exclude.list - >$$@
 package_lists += .stage4-done-all-plus-exsv.$(1).list
 .stage4-done-included.$(1).list: .stage4-done-all-plus-exsv.$(1).list .included.$(1).list
	@comm -12 .stage4-done-all-plus-exsv.$(1).list .included.$(1).list >$$@
 package_lists += .stage4-done-included.$(1).list
 .stage4-not-done-included.$(1).list: .stage4-done-all-plus-exsv.$(1).list .included.$(1).list
	@comm -13 .stage4-done-all-plus-exsv.$(1).list .included.$(1).list >$$@
 package_lists += .stage4-not-done-included.$(1).list
 .stage4-not-done-xincluded.$(1).list: .stage4-not-done-included.$(1).list .stage4-not-done-included-all.list
	@comm -23 .stage4-not-done-included.$(1).list .stage4-not-done-included-all.list >$$@
 package_lists += .stage4-not-done-xincluded.$(1).list
 list-stage4-not-done-xincluded-$(1): .stage4-not-done-xincluded.$(1).list
	@cat $$<
 phonytargets += list-stage4-not-done-xincluded-$(1)
 list-stage4-not-done-xincluded-$(1)-files: .stage4-not-done-xincluded.$(1).srpms
	@cat $$<
 phonytargets += list-stage4-not-done-xincluded-$(1)-files
 srpm_lists += .stage4-not-done-xincluded.$(1).srpms
 stage4-not-done-xincluded.$(1).srpms: .stage4-not-done-xincluded.$(1).srpms
	cat $$< >$$@
 export_srpm_lists += stage4-not-done-xincluded.$(1).srpms
endef
$(foreach arch,$(ARCHES),$(eval $(call stage4_arch_inc_template,$(arch))))

.stage4-not-done-included-all.list: $(foreach arch,$(ARCHES),.stage4-not-done-included.$(arch).list)
	@cat $(firstword $^) $(foreach f,$(wordlist 2,$(words $^),$^),| comm -12 $(f) -) | sort -u >$@
package_lists += .stage4-not-done-included-all.list

list-stage4-not-done-included-all: .stage4-not-done-included-all.list
	@cat $<
phonytargets += list-stage4-not-done-included-all

list-stage4-not-done-included-all-files: .stage4-not-done-included-all.srpms
	@cat $<
phonytargets += list-stage4-not-done-included-all-files
srpm_lists += .stage4-not-done-included-all.srpms

stage4-not-done-included.all.srpms: .stage4-not-done-included-all.srpms
	cat $< >$@
export_srpm_lists += stage4-not-done-included.all.srpms


### --- Stage V (rebuild all) ---

define stage5v_unused_arch_list_template =
 .stage5v-prerebuild.$(1).list: .stage4-done-included-all.list .stage4-done-xincluded.$(1).list
	@sort -u $$^ >$$@
 package_lists += .stage5v-prerebuild.$(1).list
 .stage5v-rebuild.$(1).list: .stage5v-prerebuild.$(1).list .stage5-prerebuild-intersection.list
	@comm -23 .stage5v-prerebuild.$(1).list .stage5-prerebuild-intersection.list >$$@
 package_lists += .stage5v-rebuild.$(1).list
 .stage5v-findlist.$(1).list: $$(REBUILD_V_DIR)
	@find -L $$(REBUILD_V_DIR) -maxdepth 1 -type d -name "*.$(1)" -printf '%f\n' | sed -e 's/\.[^.]\+$$$$//' | sort -u >$$@
 package_lists += .stage5v-findlist.$(1).list
 .stage5v-unused.$(1).list: .stage5v-rebuild.$(1).list .stage5v-findlist.$(1).list
	@comm -13 ".stage5v-rebuild.$(1).list" ".stage5v-findlist.$(1).list" >$$@
 package_lists += .stage5v-unused.$(1).list
 .stage5v-unused.$(1).dirs: .stage5v-unused.$(1).list
	@sed -e 's:^.*$$$$:$$(REBUILD_V_DIR)/&.$(1):' $$< >$$@
 package_lists += .stage5v-unused.$(1).dirs
 list-stage5v-unused-$(1)-dirs: .stage5v-unused.$(1).dirs
	@cat $$<
 phonytargets += list-stage5v-unused-$(1)-dirs
endef
$(foreach arch,$(ARCHES),$(eval $(call stage5v_unused_arch_list_template,$(arch))))

.stage5v-unused.dirs: $(foreach arch,$(ARCHES),.stage5v-unused.$(arch).dirs)
	@sort -u $^ >$@
package_lists += .stage5v-unused.dirs

list-stage5v-unused-dirs: .stage5v-unused.dirs
	@cat $<
phonytargets += list-stage5v-unused-dirs

clean-stage5v-unused: .stage5v-unused.dirs
	@cat $< | while read -r d; do rm -rfv "$$d"; done
	@rm -f .stage5v-unused.dirs $(foreach arch,$(ARCHES),.stage5v-unused.$(arch).dirs .stage5v-unused.$(arch).list)
	@rm -f .stage5v-unused.list
phonytargets += clean-stage5v-unused

define stage5v_arch_status_list_template =
 .stage5v-$(1).$(2).list: $$(STAGE5_V_STATUS) .stage5v-rebuild.$(2).list
	@$$(call filter_arch_status,$$(STAGE5_V_STATUS),$(2),REBUILD\t$(3)) | comm -12 .stage5v-rebuild.$(2).list - >$$@
 package_lists += .stage5v-$(1).$(2).list
endef
define stage5v_status_list_template =
 $$(foreach arch,$$(ARCHES),$$(eval $$(call stage5v_arch_status_list_template,$(1),$$(arch),$(2))))
endef

$(eval $(call stage5v_status_list_template,statdone,\(BLDOK\t\tCPOK\|ALRDY\)))
$(eval $(call stage5v_status_list_template,blderr,BLDERR))
$(eval $(call stage5v_status_list_template,excl,BLDERR\tEXCL))
$(eval $(call stage5v_status_list_template,brkndep,BLDERR\tBRKNDEP))
$(eval $(call stage5v_status_list_template,chkerr,BLDERR\tCHKERR))
$(eval $(call stage5v_status_list_template,nospace,BLDERR\tNOSPACE))
$(eval $(call stage5v_status_list_template,noinode,BLDERR\tNOINODE))
$(eval $(call stage5v_status_list_template,nocopy,BLDOK\t\t\(EXPERR\|CPERR\)))

define stage5v_arch_lists_template =
 .stage5v-other.$(1).list: .stage5v-blderr.$(1).list .stage5v-excl.$(1).list .stage5v-brkndep.$(1).list .stage5v-chkerr.$(1).list .stage5v-nospace.$(1).list .stage5v-noinode.$(1).list
	@cat ".stage5v-excl.$(1).list" ".stage5v-brkndep.$(1).list" ".stage5v-chkerr.$(1).list" ".stage5v-nospace.$(1).list" ".stage5v-noinode.$(1).list" | sort -u | comm -23 ".stage5v-blderr.$(1).list" - >$$@
 .stage5v-noatt.$(1).list: .stage5v-blderr.$(1).list .stage5v-done.$(1).list .stage5v-nocopy.$(1).list .stage5v-rebuild.$(1).list
	@cat ".stage5v-done.$(1).list" ".stage5v-blderr.$(1).list" ".stage5v-nocopy.$(1).list" | sort -u | comm -23 ".stage5v-rebuild.$(1).list" - >$$@
 .stage5v-miss.$(1).list: .stage5v-rebuild.$(1).list .stage5v-done.$(1).list
	@comm -23 ".stage5v-rebuild.$(1).list" ".stage5v-done.$(1).list" >$$@
 .stage5v-done.$(1).list: .stage5v-statdone.$(1).list .stage5v-findlist.$(1).list
	@comm -12 ".stage5v-statdone.$(1).list" ".stage5v-findlist.$(1).list" >$$@
 package_lists += .stage5v-other.$(1).list .stage5v-noatt.$(1).list .stage5v-miss.$(1).list .stage5v-done.$(1).list
endef
$(foreach arch,$(ARCHES),$(eval $(call stage5v_arch_lists_template,$(arch))))

retrycats5v = brkndep chkerr nospace noinode nocopy other
statcats5v = unused done excl blderr $(retrycats5v) noatt miss

$(foreach cat,$(statcats5v),$(eval $(call print_list_template,stage5v,$(cat))))
$(foreach cat,$(statcats5v),$(foreach arch,$(ARCHES),$(eval $(call print_arch_list_template,stage5v,$(cat),$(arch)))))
$(foreach cat,$(statcats5v),$(foreach arch,$(ARCHES),$(eval $(call print_arch_logs_template,stage5v,$(cat),$(arch)))))
$(foreach cat,$(statcats5v),$(eval $(call united_list_template,.stage5v-$(cat))))

stage5v-status: .stage4-done.list $(foreach cat,$(statcats5v),$(foreach arch,$(ARCHES),.stage5v-$(cat).$(arch).list)) $(foreach arch,$(ARCHES),.stage5v-miss.$(arch).list) .stage5v-unused.list .stage5v-done.list $(foreach arch,$(ARCHES),.stage5v-miss.$(arch).list)
	@echo "Total verified cleaned-up SRPM packages:  $$(cat ".stage4-done.list" | wc -l)"
	@echo
	@( \
		echo "Arch|Done|Excl|BDep|Chk|NoSI|Other|NoAtt|Total|Miss"; \
		echo "----|----|----|----|----|----|----|----|----|----"; \
		for arch in $(ARCHES); do \
			pkgs_done="$$(cat ".stage5v-done.$$arch.list" | wc -l)"; \
			pkgs_excl="$$(cat ".stage5v-excl.$$arch.list" | wc -l)"; \
			pkgs_brkndep="$$(cat ".stage5v-brkndep.$$arch.list" | wc -l)"; \
			pkgs_chkerr="$$(cat ".stage5v-chkerr.$$arch.list" | wc -l)"; \
			pkgs_nospace="$$(cat ".stage5v-nospace.$$arch.list" | wc -l)"; \
			pkgs_noinode="$$(cat ".stage5v-noinode.$$arch.list" | wc -l)"; \
			pkgs_other="$$(cat ".stage5v-other.$$arch.list" | wc -l)"; \
			pkgs_noatt="$$(cat ".stage5v-noatt.$$arch.list" | wc -l)"; \
			pkgs_miss="$$(cat ".stage5v-miss.$$arch.list" | wc -l)"; \
			echo -n "$$arch"; \
			echo -n "|$$pkgs_done"; \
			echo -n "|$$pkgs_excl"; \
			echo -n "|$$pkgs_brkndep"; \
			echo -n "|$$pkgs_chkerr"; \
			echo -n "|$$((pkgs_nospace + pkgs_noinode))"; \
			echo -n "|$$pkgs_other"; \
			echo -n "|$$pkgs_noatt"; \
			echo -n "|$$((pkgs_done + pkgs_excl + pkgs_brkndep + pkgs_chkerr + pkgs_nospace + pkgs_noinode + pkgs_other + pkgs_noatt))"; \
			echo -n "|$$pkgs_miss"; \
			echo; \
		done \
	) | column -t -s '|'
	@echo
	@echo "Total verified rebuilds missing:     $$($(if $(ARCHES),cat $(foreach arch,$(ARCHES),".stage5v-miss.$(arch).list"),echo -n) | wc -l)"
	@echo "Total unique verified SRPMs rebuilt: $$(cat ".stage5v-done.list" | wc -l)"
	@echo "Unused rebuilt verified SRPMs:       $$(cat ".stage5v-unused.list" | wc -l)"
phonytargets += stage5-verified-status

define stage5o_unused_arch_list_template =
 .stage5o-prerebuild.$(1).list: .stage4-not-done-included-all.list .stage4-not-done-xincluded.$(1).list
	@sort -u $$^ >$$@
 package_lists += .stage5o-prerebuild.$(1).list
 .stage5o-rebuild.$(1).list: .stage5o-prerebuild.$(1).list .stage5-prerebuild-intersection.list .included.$(1).list
	@sort -u .stage5o-prerebuild.$(1).list .stage5-prerebuild-intersection.list | comm -12 - .included.$(1).list >$$@
 package_lists += .stage5o-rebuild.$(1).list
 .stage5o-findlist.$(1).list: $$(REBUILD_O_DIR)
	@find -L $$(REBUILD_O_DIR) -maxdepth 1 -type d -name "*.$(1)" -printf '%f\n' | sed -e 's/\.[^.]\+$$$$//' | sort -u >$$@
 package_lists += .stage5o-findlist.$(1).list
 .stage5o-unused.$(1).list: .stage5o-rebuild.$(1).list .stage5o-findlist.$(1).list
	@comm -13 ".stage5o-rebuild.$(1).list" ".stage5o-findlist.$(1).list" >$$@
 package_lists += .stage5o-unused.$(1).list
 .stage5o-unused.$(1).dirs: .stage5o-unused.$(1).list
	@sed -e 's:^.*$$$$:$$(REBUILD_O_DIR)/&.$(1):' $$< >$$@
 package_lists += .stage5o-unused.$(1).dirs
 list-stage5o-unused-$(1)-dirs: .stage5o-unused.$(1).dirs
	@cat $$<
 phonytargets += list-stage5o-unused-$(1)-dirs
endef
$(foreach arch,$(ARCHES),$(eval $(call stage5o_unused_arch_list_template,$(arch))))

$(eval $(call united_list_template,.stage5v-prerebuild))
$(eval $(call united_list_template,.stage5o-prerebuild))

.stage5-prerebuild-intersection.list: .stage5v-prerebuild.list .stage5o-prerebuild.list
	@comm -12 $^ >$@
package_lists += .stage5-prerebuild-intersection.list

.stage5o-unused.dirs: $(foreach arch,$(ARCHES),.stage5o-unused.$(arch).dirs)
	@sort -u $^ >$@
package_lists += .stage5o-unused.dirs

list-stage5o-unused-dirs: .stage5o-unused.dirs
	@cat $<
phonytargets += list-stage5o-unused-dirs

clean-stage5o-unused: .stage5o-unused.dirs
	@cat $< | while read -r d; do rm -rfv "$$d"; done
	@rm -f .stage5o-unused.dirs $(foreach arch,$(ARCHES),.stage5o-unused.$(arch).dirs .stage5o-unused.$(arch).list)
	@rm -f .stage5o-unused.list
phonytargets += clean-stage5o-unused

define stage5o_arch_status_list_template =
 .stage5o-$(1).$(2).list: $$(STAGE5_O_STATUS) .stage5o-rebuild.$(2).list
	@$$(call filter_arch_status,$$(STAGE5_O_STATUS),$(2),REBUILD\t$(3)) | comm -12 .stage5o-rebuild.$(2).list - >$$@
 package_lists += .stage5o-$(1).$(2).list
endef
define stage5o_status_list_template =
 $$(foreach arch,$$(ARCHES),$$(eval $$(call stage5o_arch_status_list_template,$(1),$$(arch),$(2))))
endef

$(eval $(call stage5o_status_list_template,statdone,\(BLDOK\t\tCPOK\|ALRDY\)))
$(eval $(call stage5o_status_list_template,blderr,BLDERR))
$(eval $(call stage5o_status_list_template,excl,BLDERR\tEXCL))
$(eval $(call stage5o_status_list_template,brkndep,BLDERR\tBRKNDEP))
$(eval $(call stage5o_status_list_template,chkerr,BLDERR\tCHKERR))
$(eval $(call stage5o_status_list_template,nospace,BLDERR\tNOSPACE))
$(eval $(call stage5o_status_list_template,noinode,BLDERR\tNOINODE))
$(eval $(call stage5o_status_list_template,nocopy,BLDOK\t\t\(EXPERR\|CPERR\)))

define stage5o_arch_lists_template =
 .stage5o-other.$(1).list: .stage5o-blderr.$(1).list .stage5o-excl.$(1).list .stage5o-brkndep.$(1).list .stage5o-chkerr.$(1).list .stage5o-nospace.$(1).list .stage5o-noinode.$(1).list
	@cat ".stage5o-excl.$(1).list" ".stage5o-brkndep.$(1).list" ".stage5o-chkerr.$(1).list" ".stage5o-nospace.$(1).list" ".stage5o-noinode.$(1).list" | sort -u | comm -23 ".stage5o-blderr.$(1).list" - >$$@
 .stage5o-noatt.$(1).list: .stage5o-blderr.$(1).list .stage5o-done.$(1).list .stage5o-nocopy.$(1).list .stage5o-rebuild.$(1).list
	@cat ".stage5o-done.$(1).list" ".stage5o-blderr.$(1).list" ".stage5o-nocopy.$(1).list" | sort -u | comm -23 ".stage5o-rebuild.$(1).list" - >$$@
 .stage5o-miss.$(1).list: .stage5o-rebuild.$(1).list .stage5o-done.$(1).list
	@comm -23 ".stage5o-rebuild.$(1).list" ".stage5o-done.$(1).list" >$$@
 .stage5o-done.$(1).list: .stage5o-statdone.$(1).list .stage5o-findlist.$(1).list
	@comm -12 ".stage5o-statdone.$(1).list" ".stage5o-findlist.$(1).list" >$$@
 package_lists += .stage5o-other.$(1).list .stage5o-noatt.$(1).list .stage5o-miss.$(1).list .stage5o-done.$(1).list
endef
$(foreach arch,$(ARCHES),$(eval $(call stage5o_arch_lists_template,$(arch))))

retrycats5o = brkndep chkerr nospace noinode nocopy other
statcats5o = unused done excl blderr $(retrycats5o) noatt miss

$(foreach cat,$(statcats5o),$(eval $(call print_list_template,stage5o,$(cat))))
$(foreach cat,$(statcats5o),$(foreach arch,$(ARCHES),$(eval $(call print_arch_list_template,stage5o,$(cat),$(arch)))))
$(foreach cat,$(statcats5o),$(foreach arch,$(ARCHES),$(eval $(call print_arch_logs_template,stage5o,$(cat),$(arch)))))
$(foreach cat,$(statcats5o),$(eval $(call united_list_template,.stage5o-$(cat))))

stage5o-status: .stage4-not-done-included.list $(foreach cat,$(statcats5o),$(foreach arch,$(ARCHES),.stage5o-$(cat).$(arch).list)) $(foreach arch,$(ARCHES),.stage5o-miss.$(arch).list) .stage5o-unused.list .stage5o-done.list $(foreach arch,$(ARCHES),.stage5o-miss.$(arch).list)
	@echo "Total not verified SRPM packages:  $$(cat ".stage4-not-done-included.list" | wc -l)"
	@echo
	@( \
		echo "Arch|Done|Excl|BDep|Chk|NoSI|Other|NoAtt|Total|Miss"; \
		echo "----|----|----|----|----|----|----|----|----|----"; \
		for arch in $(ARCHES); do \
			pkgs_done="$$(cat ".stage5o-done.$$arch.list" | wc -l)"; \
			pkgs_excl="$$(cat ".stage5o-excl.$$arch.list" | wc -l)"; \
			pkgs_brkndep="$$(cat ".stage5o-brkndep.$$arch.list" | wc -l)"; \
			pkgs_chkerr="$$(cat ".stage5o-chkerr.$$arch.list" | wc -l)"; \
			pkgs_nospace="$$(cat ".stage5o-nospace.$$arch.list" | wc -l)"; \
			pkgs_noinode="$$(cat ".stage5o-noinode.$$arch.list" | wc -l)"; \
			pkgs_other="$$(cat ".stage5o-other.$$arch.list" | wc -l)"; \
			pkgs_noatt="$$(cat ".stage5o-noatt.$$arch.list" | wc -l)"; \
			pkgs_miss="$$(cat ".stage5o-miss.$$arch.list" | wc -l)"; \
			echo -n "$$arch"; \
			echo -n "|$$pkgs_done"; \
			echo -n "|$$pkgs_excl"; \
			echo -n "|$$pkgs_brkndep"; \
			echo -n "|$$pkgs_chkerr"; \
			echo -n "|$$((pkgs_nospace + pkgs_noinode))"; \
			echo -n "|$$pkgs_other"; \
			echo -n "|$$pkgs_noatt"; \
			echo -n "|$$((pkgs_done + pkgs_excl + pkgs_brkndep + pkgs_chkerr + pkgs_nospace + pkgs_noinode + pkgs_other + pkgs_noatt))"; \
			echo -n "|$$pkgs_miss"; \
			echo; \
		done \
	) | column -t -s '|'
	@echo
	@echo "Total original rebuilds missing:     $$($(if $(ARCHES),cat $(foreach arch,$(ARCHES),".stage5o-miss.$(arch).list"),echo -n) | wc -l)"
	@echo "Total unique original SRPMs rebuilt: $$(cat ".stage5o-done.list" | wc -l)"
	@echo "Unused rebuilt original SRPMs:       $$(cat ".stage5o-unused.list" | wc -l)"
phonytargets += stage5-verified-status

$(STAGE5_V_STATUS):
	@touch $@

ifneq ($(dir $(STAGE5_V_LOG)),$(dir $(STAGE4_LOG)))
$(dir $(STAGE5_V_LOG)):
	mkdir -p $@
endif

$(STAGE5_V_LOGDIR):
	mkdir -p $@

stage5v: stage5v-allarch $(foreach arch,$(ARCHES),stage5v-$(arch))
phonytargets += stage5v

stage5v-dry-run: stage5v-allarch-dry-run $(foreach arch,$(ARCHES),stage5v-$(arch)-dry-run)
phonytargets += stage5v-dry-run

stage5v-dry-run-run: stage5v-allarch-dry-run-run $(foreach arch,$(ARCHES),stage5v-$(arch)-dry-run-run)
phonytargets += stage5v-dry-run-run

stage5v-rebuild-all: stage5v-rebuild-allarch $(foreach arch,$(ARCHES),stage5v-rebuild-x-$(arch))
phonytargets += stage5v-rebuild-all

.stage5v-noatt-all.srpms: .stage4-done-included-all.srpms $(foreach arch,$(ARCHES),.stage5v-noatt.$(arch).srpms)
	@cat $(firstword $^) $(foreach f,$(wordlist 2,$(words $^),$^),| comm -12 $(f) -) | sort -u >$@
srpm_lists += .stage5v-noatt-all.srpms

define rebuild_verified_included_all_template
 stage5v$(1): $(2) $$(if $$(DISTTAG_OPTS),$$(DISTTAG_INDEX)) | $$(SRPM_DIR) $$(dir $$(STAGE5_V_LOG)) $$(STAGE5_V_LOGDIR)
	$$(call backup_log,$$(STAGE5_V_LOG)$$(LOG_SUF))
	-time $$(HSH_SEPARATE_SOURCES) $$(BUILD_COMMON_OPTS) --nodetect -O "$$(REBUILD_V_DIR)" -l "$$(STAGE5_V_LOGDIR)" $$(DISTTAG_OPTS) $$(OPTS) $$(REBUILD_OPTS) -B "$$(SRPM_DIR)" --with-srpms="$$(SRPM_DIR)" $$(if $$(FOR_ARCHES),$$(call STAGE4_BUILDERS,$$(subst $$(comma), ,$$(FOR_ARCHES))),$$(STAGE4_BUILDERS)) $$< 2>&1 $$(if $$(DRY_RUN),,| tee "$$(STAGE5_V_LOG)$$(LOG_SUF)")
	$$(call import_log,$$(STAGE5_V_LOG)$$(LOG_SUF),$$(STAGE5_V_STATUS))
	touch "$$(REBUILD_V_DIR)"
 phonytargets += stage5v$(1)
 stage5v$(1)-dry-run:
	@$$(MAKE_SELF) stage5v$(1) DRY_RUN=1 OPTS="--dry-run $$(OPTS)" FOR_ARCHES="$$(FOR_ARCHES)"
 phonytargets += stage5v$(1)-dry-run
 stage5v$(1)-dry-run-run:
	@$$(MAKE_SELF) stage5v$(1) DRY_RUN=1 OPTS="--dry-run-run $$(OPTS)" FOR_ARCHES="$$(FOR_ARCHES)"
 phonytargets += stage5v$(1)-dry-run-run
 srpm_lists += $(2)
endef

$(eval $(call rebuild_verified_included_all_template,-rebuild-allarch,.stage4-done-included-all.srpms))
$(eval $(call rebuild_verified_included_all_template,-allarch,.stage5v-noatt-all.srpms))

define rebuild_verified_included_arch_template
 stage5v-$(1)$(2)$(3): $(4) $$(if $$(DISTTAG_OPTS),$$(DISTTAG_INDEX)) | $$(SRPM_DIR) $$(dir $$(STAGE5_V_LOG)) $$(STAGE5_V_LOGDIR)
	$$(call backup_log,$$(STAGE5_V_LOG)$$(LOG_SUF))
	-time $$(HSH_SEPARATE_SOURCES) $$(BUILD_COMMON_OPTS) --nodetect -O "$$(REBUILD_V_DIR)" -l "$$(STAGE5_V_LOGDIR)" $$(DISTTAG_OPTS) $$(OPTS) $$(REBUILD_OPTS) -B "$$(SRPM_DIR)" --with-srpms="$$(SRPM_DIR)" $$(call STAGE4_BUILDERS,$(3)) $$< 2>&1 $$(if $$(DRY_RUN),,| tee "$$(STAGE5_V_LOG)$$(LOG_SUF)")
	$$(call import_log,$$(STAGE5_V_LOG)$$(LOG_SUF),$$(STAGE5_V_STATUS))
	touch "$$(REBUILD_V_DIR)"
 phonytargets += stage5v-$(1)$(2)$(3)
 stage5v-$(1)$(2)$(3)-dry-run:
	@$$(MAKE_SELF) stage5v-$(1)$(2)$(3) DRY_RUN=1 OPTS="--dry-run $$(OPTS)"
 phonytargets += stage5v-$(1)$(2)$(3)-dry-run
 stage5v-$(1)$(2)$(3)-dry-run-run:
	@$$(MAKE_SELF) stage5v-$(1)$(2)$(3) DRY_RUN=1 OPTS="--dry-run-run $$(OPTS)"
 phonytargets += stage5v-$(1)$(2)$(3)-dry-run-run
 srpm_lists += $(4)
endef
$(foreach arch,$(ARCHES),$(eval $(call rebuild_verified_included_arch_template,rebuild-x-,,$(arch),.stage4-done-xincluded.$(arch).srpms)))
$(foreach cat,$(retrycats5v),$(foreach arch,$(ARCHES),$(eval $(call rebuild_verified_included_arch_template,retry-,$(cat)-,$(arch),.stage5v-$(cat).$(arch).srpms))))
$(foreach arch,$(ARCHES),$(eval $(call rebuild_verified_included_arch_template,,,$(arch),.stage5v-noatt.$(arch).srpms)))

define rebuild_verified_included_template
 stage5v-$(1)$(2): $$(foreach arch,$$(if $$(FOR_ARCHES),$$(filter $$(subst $$(comma), ,$$(FOR_ARCHES)),$$(ARCHES)),$$(ARCHES)),stage5v-$(1)$(2)-$$(arch))
 phonytargets += stage5v-$(1)$(2)
 stage5v-$(1)$(2)-dry-run:
	@$$(MAKE_SELF) stage5v-$(1)$(2) DRY_RUN=1 OPTS="--dry-run $$(OPTS)" FOR_ARCHES="$$(FOR_ARCHES)"
 phonytargets += stage5v-$(1)$(2)-dry-run
 stage5v-$(1)$(2)-dry-run-run:
	@$$(MAKE_SELF) stage5v-$(1)$(2) DRY_RUN=1 OPTS="--dry-run-run $$(OPTS)" FOR_ARCHES="$$(FOR_ARCHES)"
 phonytargets += stage5v-$(1)$(2)-dry-run-run
endef
$(foreach cat,$(retrycats5v),$(eval $(call rebuild_verified_included_template,retry-,$(cat))))

stage5v-retry-pkgs: $(if $(DISTTAG_OPTS),$(DISTTAG_INDEX)) | $(SRPM_DIR) $(dir $(STAGE5_V_LOG)) $(STAGE5_V_LOGDIR)
	$(call backup_log,$(STAGE5_V_LOG)$(LOG_SUF))
	-time $(HSH_SEPARATE_SOURCES) $(BUILD_COMMON_OPTS) --nodetect -O "$(REBUILD_V_DIR)" -l "$(STAGE5_V_LOGDIR)" $(DISTTAG_OPTS) $(OPTS) $(REBUILD_OPTS) -B "$(SRPM_DIR)" --with-srpms="$(SRPM_DIR)" $(if $(FOR_ARCHES),$(call STAGE4_BUILDERS,$(subst $(comma), ,$(FOR_ARCHES))),$(STAGE4_BUILDERS)) $(addsuffix .src.rpm,$(PKGS)) 2>&1 $(if $(DRY_RUN),,| tee "$(STAGE5_V_LOG)$(LOG_SUF)")
	$(call import_log,$(STAGE5_V_LOG)$(LOG_SUF),$(STAGE5_V_STATUS))
	touch "$(REBUILD_V_DIR)"
phonytargets += stage5v-retry-pkgs
stage5v-retry-pkgs-dry-run:
	@$(MAKE_SELF) stage5v-retry-pkgs DRY_RUN=1 OPTS="--dry-run $(OPTS)" PKGS="$(PKGS)" FOR_ARCHES="$(FOR_ARCHES)"
phonytargets += stage5v-retry-pkgs-dry-run
stage5v-retry-pkgs-dry-run-run:
	@$(MAKE_SELF) stage5v-retry-pkgs DRY_RUN=1 OPTS="--dry-run-run $(OPTS)" PKGS="$(PKGS)" FOR_ARCHES="$(FOR_ARCHES)"
phonytargets += stage5v-retry-pkgs-dry-run-run

$(STAGE5_O_STATUS):
	@touch $@

ifneq ($(dir $(STAGE5_V_LOG)),$(dir $(STAGE5_O_LOG)))
$(dir $(STAGE5_O_LOG)):
	mkdir -p $@
endif

$(STAGE5_O_LOGDIR):
	mkdir -p $@

stage5o: stage5o-allarch $(foreach arch,$(ARCHES),stage5o-$(arch))
phonytargets += stage5o

stage5o-dry-run: stage5o-allarch-dry-run $(foreach arch,$(ARCHES),stage5o-$(arch)-dry-run)
phonytargets += stage5o-dry-run

stage5o-dry-run-run: stage5o-allarch-dry-run-run $(foreach arch,$(ARCHES),stage5o-$(arch)-dry-run-run)
phonytargets += stage5o-dry-run-run

stage5o-rebuild-all: stage5o-rebuild-allarch $(foreach arch,$(ARCHES),stage5o-rebuild-x-$(arch))
phonytargets += stage5o-rebuild-all

.stage5o-noatt-all.srpms: .stage4-not-done-included-all.srpms $(foreach arch,$(ARCHES),.stage5o-noatt.$(arch).srpms)
	@cat $(firstword $^) $(foreach f,$(wordlist 2,$(words $^),$^),| comm -12 $(f) -) | sort -u >$@
srpm_lists += .stage5o-noatt-all.srpms

define rebuild_orig_included_all_template
 stage5o$(1): $(2) $$(if $$(DISTTAG_OPTS),$$(DISTTAG_INDEX)) | $$(dir $$(STAGE5_O_LOG)) $$(STAGE5_O_LOGDIR)
	$$(call backup_log,$$(STAGE5_O_LOG)$$(LOG_SUF))
	-time $$(HSH_SEPARATE_SOURCES) $$(BUILD_COMMON_OPTS) --nodetect -O "$$(REBUILD_O_DIR)" -l "$$(STAGE5_O_LOGDIR)" $$(DISTTAG_OPTS) $$(OPTS) $$(REBUILD_OPTS) $$(if $$(FOR_ARCHES),$$(call BUILDERS,$$(subst $$(comma), ,$$(FOR_ARCHES))),$$(BUILDERS)) $$< 2>&1 $$(if $$(DRY_RUN),,| tee "$$(STAGE5_O_LOG)$$(LOG_SUF)")
	$$(call import_log,$$(STAGE5_O_LOG)$$(LOG_SUF),$$(STAGE5_O_STATUS))
	touch "$$(REBUILD_O_DIR)"
 phonytargets += stage5o$(1)
 stage5o$(1)-dry-run:
	@$$(MAKE_SELF) stage5o$(1) DRY_RUN=1 OPTS="--dry-run $$(OPTS)" FOR_ARCHES="$$(FOR_ARCHES)"
 phonytargets += stage5o$(1)-dry-run
 stage5o$(1)-dry-run-run:
	@$$(MAKE_SELF) stage5o$(1) DRY_RUN=1 OPTS="--dry-run-run $$(OPTS)" FOR_ARCHES="$$(FOR_ARCHES)"
 phonytargets += stage5o$(1)-dry-run-run
 srpm_lists += $(2)
endef

$(eval $(call rebuild_orig_included_all_template,-rebuild-allarch,.stage4-not-done-included-all.srpms))
$(eval $(call rebuild_orig_included_all_template,-allarch,.stage5o-noatt-all.srpms))

define rebuild_orig_included_arch_template
 stage5o-$(1)$(2)$(3): $(4) $$(if $$(DISTTAG_OPTS),$$(DISTTAG_INDEX)) | $$(dir $$(STAGE5_O_LOG)) $$(STAGE5_O_LOGDIR)
	$$(call backup_log,$$(STAGE5_O_LOG)$$(LOG_SUF))
	-time $$(HSH_SEPARATE_SOURCES) $$(BUILD_COMMON_OPTS) --nodetect -O "$$(REBUILD_O_DIR)" -l "$$(STAGE5_O_LOGDIR)" $$(DISTTAG_OPTS) $$(OPTS) $$(REBUILD_OPTS) $$(call BUILDERS,$(3)) $$< 2>&1 $$(if $$(DRY_RUN),,| tee "$$(STAGE5_O_LOG)$$(LOG_SUF)")
	$$(call import_log,$$(STAGE5_O_LOG)$$(LOG_SUF),$$(STAGE5_O_STATUS))
	touch "$$(REBUILD_O_DIR)"
 phonytargets += stage5o-$(1)$(2)$(3)
 stage5o-$(1)$(2)$(3)-dry-run:
	@$$(MAKE_SELF) stage5o-$(1)$(2)$(3) DRY_RUN=1 OPTS="--dry-run $$(OPTS)"
 phonytargets += stage5o-$(1)$(2)$(3)-dry-run
 stage5o-$(1)$(2)$(3)-dry-run-run:
	@$$(MAKE_SELF) stage5o-$(1)$(2)$(3) DRY_RUN=1 OPTS="--dry-run-run $$(OPTS)"
 phonytargets += stage5o-$(1)$(2)$(3)-dry-run-run
 srpm_lists += $(4)
endef
$(foreach arch,$(ARCHES),$(eval $(call rebuild_orig_included_arch_template,rebuild-x-,,$(arch),.stage4-not-done-xincluded.$(arch).srpms)))
$(foreach cat,$(retrycats5o),$(foreach arch,$(ARCHES),$(eval $(call rebuild_orig_included_arch_template,retry-,$(cat)-,$(arch),.stage5o-$(cat).$(arch).srpms))))
$(foreach arch,$(ARCHES),$(eval $(call rebuild_orig_included_arch_template,,,$(arch),.stage5o-noatt.$(arch).srpms)))

define rebuild_orig_included_template
 stage5o-$(1)$(2): $$(foreach arch,$$(if $$(FOR_ARCHES),$$(filter $$(subst $$(comma), ,$$(FOR_ARCHES)),$$(ARCHES)),$$(ARCHES)),stage5o-$(1)$(2)-$$(arch))
 phonytargets += stage5o-$(1)$(2)
 stage5o-$(1)$(2)-dry-run:
	@$$(MAKE_SELF) stage5o-$(1)$(2) DRY_RUN=1 OPTS="--dry-run $$(OPTS)" FOR_ARCHES="$$(FOR_ARCHES)"
 phonytargets += stage5o-$(1)$(2)-dry-run
 stage5o-$(1)$(2)-dry-run-run:
	@$$(MAKE_SELF) stage5o-$(1)$(2) DRY_RUN=1 OPTS="--dry-run-run $$(OPTS)" FOR_ARCHES="$$(FOR_ARCHES)"
 phonytargets += stage5o-$(1)$(2)-dry-run-run
endef
$(foreach cat,$(retrycats5o),$(eval $(call rebuild_orig_included_template,retry-,$(cat))))

stage5o-retry-pkgs: $(if $(DISTTAG_OPTS),$(DISTTAG_INDEX)) | $(dir $(STAGE5_O_LOG)) $(STAGE5_O_LOGDIR)
	$(call backup_log,$(STAGE5_O_LOG)$(LOG_SUF))
	-time $(HSH_SEPARATE_SOURCES) $(BUILD_COMMON_OPTS) --nodetect -O "$(REBUILD_O_DIR)" -l "$(STAGE5_O_LOGDIR)" $(DISTTAG_OPTS) $(OPTS) $(REBUILD_OPTS) $(if $(FOR_ARCHES),$(call BUILDERS,$(subst $(comma), ,$(FOR_ARCHES))),$(BUILDERS)) $(addsuffix .src.rpm,$(PKGS)) 2>&1 $(if $(DRY_RUN),,| tee "$(STAGE5_O_LOG)$(LOG_SUF)")
	$(call import_log,$(STAGE5_O_LOG)$(LOG_SUF),$(STAGE5_O_STATUS))
	touch "$(REBUILD_O_DIR)"
phonytargets += stage5o-retry-pkgs
stage5o-retry-pkgs-dry-run:
	@$(MAKE_SELF) stage5o-retry-pkgs DRY_RUN=1 OPTS="--dry-run $(OPTS)" PKGS="$(PKGS)" FOR_ARCHES="$(FOR_ARCHES)"
phonytargets += stage5o-retry-pkgs-dry-run
stage5o-retry-pkgs-dry-run-run:
	@$(MAKE_SELF) stage5o-retry-pkgs DRY_RUN=1 OPTS="--dry-run-run $(OPTS)" PKGS="$(PKGS)" FOR_ARCHES="$(FOR_ARCHES)"
phonytargets += stage5o-retry-pkgs-dry-run-run


### Indices

indices =

define arch_index_template =
 $(1).v.index: $$(REBUILD_V_DIR)
	find -L '$$(REBUILD_V_DIR)' -name '*.$(1).rpm' | grep -v -e '-debuginfo-[^-]\+-[^-]\+\.$(1)\.rpm$$$$' | sort >$$@
 indices += $(1).v.index
 $(1).v-dbg.index: $$(REBUILD_V_DIR)
	find -L '$$(REBUILD_V_DIR)' -name '*.$(1).rpm' | grep -e '-debuginfo-[^-]\+-[^-]\+\.$(1)\.rpm$$$$' | sort >$$@
 indices += $(1).v-dbg.index
 $(1).o.index: $$(REBUILD_O_DIR)
	find -L '$$(REBUILD_O_DIR)' -name '*.$(1).rpm' | grep -v -e '-debuginfo-[^-]\+-[^-]\+\.$(1)\.rpm$$$$' | sort >$$@
 indices += $(1).o.index
 $(1).o-dbg.index: $$(REBUILD_O_DIR)
	find -L '$$(REBUILD_O_DIR)' -name '*.$(1).rpm' | grep -e '-debuginfo-[^-]\+-[^-]\+\.$(1)\.rpm$$$$' | sort >$$@
 indices += $(1).o-dbg.index
endef
$(foreach arch,$(ARCHES) noarch,$(eval $(call arch_index_template,$(arch))))

SRPM.v.index: $(REBUILD_V_DIR)
	find -L '$(REBUILD_V_DIR)' -name '*.src.rpm' | sort -u >$@
indices += SRPM.v.index

SRPM.o.index: $(REBUILD_O_DIR)
	find -L '$(REBUILD_O_DIR)' -name '*.src.rpm' | sort -u >$@
indices += SRPM.o.index

SRPM.index: $(foreach arch,$(ARCHES) noarch,SRPM.$(arch).v.index SRPM.$(arch).o.index)
	sort -u $^ >$@
indices += SRPM.index


### Export

export-rpms: $(indices) $(REBUILD_V_DIR) $(REBUILD_O_DIR)
	@if [ -z "$(DESTDIR)" ]; then \
		echo "Please specify a DESTDIR." >&2; \
		exit 1; \
	fi
	rsync -a --delete $(EXPORT_OPTS) "$(REBUILD_V_DIR)/" "$(DESTDIR)/$(REBUILD_V_DIR)/"
	rsync -a --delete $(EXPORT_OPTS) "$(REBUILD_O_DIR)/" "$(DESTDIR)/$(REBUILD_O_DIR)/"
	rm -f "$(DESTDIR)/"*.index "$(DESTDIR)/"*.list
	rsync -a $(indices) $(EXPORT_OPTS) "$(DESTDIR)/"
phonytargets += export-rpms

SRPM.%.index: %.index %-dbg.index
	sed -e 's,/[^/]*$$,,' -e 's,/[^/]*$$,&&,' -e 's/\.[^.]*$$/.src.rpm/' $^ | sort -u >$@

define make_arch_repo_template
 $$(OUTREPO)/$(1)/base: $(1).v.index $(1).v-dbg.index $(1).o.index $(1).o-dbg.index SRPM.$(1).v.index SRPM.$(1).o.index
	rm -rf "$$(OUTREPO)/$(1)"
	mkdir -p "$$(OUTREPO)/$(1)/RPMS.clean"
	xargs -i ln -sf ../../../{} $$(OUTREPO)/$(1)/RPMS.clean/ <$(1).v.index
	mkdir -p "$$(OUTREPO)/$(1)/RPMS.clean-dbg"
	xargs -i ln -sf ../../../{} $$(OUTREPO)/$(1)/RPMS.clean-dbg/ <$(1).v-dbg.index
	mkdir -p "$$(OUTREPO)/$(1)/SRPMS.clean"
	xargs -i ln -sf ../../../{} $$(OUTREPO)/$(1)/SRPMS.clean/ <SRPM.$(1).v.index
	mkdir -p "$$(OUTREPO)/$(1)/RPMS.orig"
	xargs -i ln -sf ../../../{} $$(OUTREPO)/$(1)/RPMS.orig/ <$(1).o.index
	mkdir -p "$$(OUTREPO)/$(1)/RPMS.orig-dbg"
	xargs -i ln -sf ../../../{} $$(OUTREPO)/$(1)/RPMS.orig-dbg/ <$(1).o-dbg.index
	mkdir -p "$$(OUTREPO)/$(1)/SRPMS.orig"
	xargs -i ln -sf ../../../{} $$(OUTREPO)/$(1)/SRPMS.orig/ <SRPM.$(1).o.index
	genbasedir --create --progress --topdir="$$(OUTREPO)" $(1) clean orig clean-dbg orig-dbg
 repo-$(1): $$(OUTREPO)/$(1)/base
 phonytargets += repo-$(1)
 indices += SRPM.$(1).v.index SRPM.$(1).o.index
endef
$(foreach arch,$(ARCHES) noarch,$(eval $(call make_arch_repo_template,$(arch))))

repo: $(foreach arch,$(ARCHES) noarch,repo-$(arch))
phonytargets += repo

export: export-rpms
phonytargets += export

export-repo: repo
	@if [ -z "$(DESTDIR)" ]; then \
		echo "Please specify a DESTDIR." >&2; \
		exit 1; \
	fi
	rsync -a --delete $(EXPORT_REPO_OPTS) "$(OUTREPO)/" "$(DESTDIR)/$(OUTREPO)/"
phonytargets += export-repo


### Disttag

# zzuf    0.15-alt1_10    ppc64le zzuf-0.15-alt1_10.ppc64le.rpm   zzuf-0.15-alt1_10.src.rpm
$(DISTTAG_INDEX): $(ALL_SRPMS) $(BIN_LIST) $(CUSTOM_TAGS)
	@echo "Rebuilding the dist-tag index..." >&2
	@cat "$(ALL_SRPMS)" | while read -r pkg; do \
		binpkg="$$(sed -n -e "/^[^[:space:]]\\+[[:space:]]\\+[^[:space:]]\\+[[:space:]]\\+[^[:space:]]\\+[[:space:]]\\+[^[:space:]]\\+[[:space:]]\\+$${pkg}\$$/ { s/^[^[:space:]]\\+[[:space:]]\\+[^[:space:]]\\+[[:space:]]\\+\\([^[:space:]]\\+\\)[[:space:]]\\+\\([^[:space:]]\\+\\)[[:space:]]\\+$${pkg}\$$/\\1\/RPMS\/\\2/; p; q }" "$(BIN_LIST)")"; \
		if [ -z "$$binpkg" ]; then \
			if [ -n '$(CUSTOM_TAGS)' -a -e '$(CUSTOM_TAGS)' ]; then \
				if grep -q "^$${pkg%.src.rpm}[[:space:]]" '$(CUSTOM_TAGS)'; then \
					grep "^$${pkg%.src.rpm}[[:space:]]" '$(CUSTOM_TAGS)'; \
					continue; \
				fi; \
			fi; \
			echo "Could not find binary package for $$pkg" >&2; \
			exit 1; \
		fi; \
		disttag="$$(rpm -q --qf '%{disttag}\n' -p "$(REPO_DIR)$(REPO_FILES_SUBDIR)/$$binpkg")"; \
		[ -n "$$disttag" ] || disttag="%nil"; \
		echo "$${pkg%.src.rpm}	$$disttag"; \
	done >$@


### Checks

check-rebuild-lists: .stage4-done-included-all.srpms .stage4-not-done-included-all.srpms $(foreach arch,$(ARCHES),.stage4-done-xincluded.$(arch).srpms .stage4-not-done-xincluded.$(arch).srpms .included.$(arch).srpms) .stage5v-unused.dirs
	@echo "* Lists:"
	@wc -l .stage4-done-included-all.srpms $(foreach arch,$(ARCHES),.stage4-done-xincluded.$(arch).srpms) .stage4-not-done-included-all.srpms $(foreach arch,$(ARCHES),.stage4-not-done-xincluded.$(arch).srpms)
	@echo
	@echo "* Intersections: $$((comm -12 .stage4-done-included-all.srpms .stage4-not-done-included-all.srpms; $(foreach arch,$(ARCHES),comm -12 .stage4-done-included-all.srpms .stage4-done-xincluded.$(arch).srpms;) $(foreach arch,$(ARCHES),comm -12 .stage4-not-done-included-all.srpms .stage4-not-done-xincluded.$(arch).srpms;) $(foreach arch,$(ARCHES),comm -12 .stage4-done-included-all.srpms .stage4-not-done-xincluded.$(arch).srpms;) $(foreach arch,$(ARCHES),comm -12 .stage4-not-done-included-all.srpms .stage4-done-xincluded.$(arch).srpms;) $(foreach arch,$(ARCHES),comm -12 .stage4-done-xincluded.$(arch).srpms .stage4-not-done-xincluded.$(arch).srpms;)) | wc -l)"
	@echo
	@echo "* Coverage:"
	@$(foreach arch,$(ARCHES),echo "    $(arch): $$(cat .included.$(arch).srpms |  wc -l)/$$(sort .stage4-done-included-all.srpms .stage4-done-xincluded.$(arch).srpms .stage4-not-done-included-all.srpms .stage4-not-done-xincluded.$(arch).srpms | comm -12 .included.$(arch).srpms - | wc -l)";)
	@echo
	@echo "* Unused list intersections: $$((sed -e 's/\.src\.rpm$$//' -e 's:^.*$$:$(foreach arch,$(ARCHES),$(REBUILD_V_DIR)/&.$(arch)\n):' -e 's/\n /\n/g' -e 's/\n$$//' .stage4-done-included-all.srpms; $(foreach arch,$(ARCHES),sed -e 's/\.src\.rpm$$//' -e 's:^.*$$:$(REBUILD_V_DIR)/&.$(arch):' .stage4-done-xincluded.$(arch).srpms;)) | sort | comm -12 .stage5v-unused.dirs - | wc -l) $$((sed -e 's/\.src\.rpm$$//' -e 's:^.*$$:$(foreach arch,$(ARCHES),$(REBUILD_O_DIR)/&.$(arch)\n):' -e 's/\n /\n/g' -e 's/\n$$//' .stage4-not-done-included-all.srpms; $(foreach arch,$(ARCHES),sed -e 's/\.src\.rpm$$//' -e 's:^.*$$:$(REBUILD_O_DIR)/&.$(arch):' .stage4-not-done-xincluded.$(arch).srpms;)) | sort | comm -12 .stage5v-unused.dirs - | wc -l)"

phonytargets += check-rebuild-lists
srpm_lists += $(foreach arch,$(ARCHES),.included.$(arch).srpms)

define check_rebuild_arch_results =
 .rebuild-orig-check.$(1).list: $$(REBUILD_O_DIR)
	@find -L $$(REBUILD_O_DIR) -maxdepth 1 -type d -name '*.$(1)' -printf '%f\n' | sed -e 's/\.$(arch)$$$$//' | sort -u >$$@
 package_lists += .rebuild-orig-check.$(1).list
 .rebuild-verified-check.$(1).list: $$(REBUILD_V_DIR)
	@find -L $$(REBUILD_V_DIR) -maxdepth 1 -type d -name '*.$(1)' -printf '%f\n' | sed -e 's/\.$(arch)$$$$//' | sort -u >$$@
 package_lists += .rebuild-verified-check.$(1).list
 check-rebuild-results-$(1): .rebuild-orig-check.$(1).list .rebuild-verified-check.$(1).list .included.$(1).srpms .stage5v-rebuild.$(1).list .stage5o-rebuild.$(1).list .stage5o-excl.$(1).list .stage5v-excl.$(1).list
	@echo "V plan/result: $$$$(cat .stage5v-rebuild.$(1).list | wc -l)/$$$$(cat .rebuild-verified-check.$(1).list | wc -l)"
	@echo "O plan/result: $$$$(cat .stage5o-rebuild.$(1).list | wc -l)/$$$$(cat .rebuild-orig-check.$(1).list | wc -l)"
	@echo "V/O intersections: $$$$(comm -12 .rebuild-orig-check.$(1).list .rebuild-verified-check.$(1).list | wc -l)"
	@echo "Coverage:    $$$$(cat .included.$(1).srpms | wc -l )/$$$$(cat .rebuild-orig-check.$(1).list .rebuild-verified-check.$(1).list | sed -e 's/^.*$$$$/&.src.rpm/' | sort | comm -12 .included.$(1).srpms - | wc -l)"
	@echo " with excl.: $$$$(cat .included.$(1).srpms | wc -l )/$$$$(cat .rebuild-orig-check.$(1).list .rebuild-verified-check.$(1).list .stage5o-excl.$(1).list .stage5v-excl.$(1).list | sed -e 's/^.*$$$$/&.src.rpm/' | sort | comm -12 .included.$(1).srpms - | wc -l)"
 phonytargets += check-rebuild-results-$(1)
endef
$(foreach arch,$(ARCHES),$(eval $(call check_rebuild_arch_results,$(arch))))

check-versions: check-included $(if $(DISTTAG_OPTS),check-disttags)
phonytargets += check-versions

check-included: $(ALL_SRPMS) .included.srpms
	-diff "$(ALL_SRPMS)" .included.srpms
phonytargets += check-included
srpm_lists += .included.srpms

# After the final indices += ...
index-rebuild: clean-indices $(indices)
phonytargets += index-rebuild

.rebuilt-disttag.list bin-disttag.list: $(REBUILD_V_DIR) $(REBUILD_O_DIR)
	@echo 'Rebuilding dist-tag indicies...' >&2
	@rm -f .rebuilt-disttag.list bin-disttag.list
	@find "$(REBUILD_V_DIR)" "$(REBUILD_O_DIR)" -mindepth 2 | while read -r path; do \
		case "$$path" in *.src.rpm) continue;; esac; \
		rpm="$${path##*/}"; \
		rpm -q --qf "%{name}\\t%{evr}\\t%{disttag}\\t%{arch}\\t$$rpm\\t%{sourcerpm}\\n" -p "$$path" >>.rebuilt-disttag.list; \
		arch="$${rpm%.rpm}"; arch="$${arch##*.}"; \
		rpm -q --qf "%{name}\\t%{evr}\\t%{disttag}\\t%{arch}\\t$$rpm\\t%{sourcerpm}\\n" -p "$(REPO_DIR)$(REPO_FILES_SUBDIR)/$$arch/RPMS/$${path##*/}" >>bin-disttag.list; \
	done
package_lists += .rebuilt-disttag.list bin-disttag.list

check-disttags: bin-disttag.list .rebuilt-disttag.list
	-diff $^
phonytargets = += check-disttags


### Status file rules

show-stage1-pkg-status:
	@$(call show_pkg_status,$(STAGE1_STATUS),$(PKG),$(ARCH))
phonytargets += show-stage1-pkg-status
show-stage2-pkg-status:
	@$(call show_pkg_status,$(STAGE2_STATUS),$(PKG))
phonytargets += show-stage2-pkg-status
show-stage3-pkg-status:
	@$(call show_pkg_status,$(STAGE3_STATUS),$(PKG))
phonytargets += show-stage3-pkg-status
show-stage4-pkg-status:
	@$(call show_pkg_status,$(STAGE4_STATUS),$(PKG),$(ARCH))
phonytargets += show-stage4-pkg-status
show-stage5v-pkg-status:
	@$(call show_pkg_status,$(STAGE5_V_STATUS),$(PKG),$(ARCH))
phonytargets += show-stage5v-pkg-status
show-stage5o-pkg-status:
	@$(call show_pkg_status,$(STAGE5_O_STATUS),$(PKG),$(ARCH))
phonytargets += show-stage5o-pkg-status


### Multistage rebuild rules

define check_pkgs_template =
	@echo 'Checking Stage $(patsubst %:$(1),%,$(filter %:$(1),I:STAGE1_STATUS II:STAGE2_STATUS III:STAGE3_STATUS IV:STAGE4_STATUS V/1:STAGE5_V_STATUS V/2:STAGE5_O_STATUS)) status for $(PKGS)...'
	@for pkg in $(PKGS); do \
		if $(call show_pkg_status,$($(1)),$$pkg) | tail -1 | \
			grep -v $(foreach status,$(2),-e '$(status)'); \
		then \
			exit 1; \
		fi; \
	done
endef

reclean-pkgs:
	$(MAKE_SELF) stage2-retry-pkgs OPTS="-f $(OPTS)" PKGS="$(PKGS)"
	@$(call check_pkgs_template,STAGE2_STATUS,TAROK NOCHGS)
	$(MAKE_SELF) stage3-retry-pkgs OPTS="$(OPTS)" PKGS="$(PKGS)"
	@$(call check_pkgs_template,STAGE3_STATUS,BLDOK[[:space:]]\+CPOK)
phonytargets += reclean-pkgs

reclean-and-rebuild-pkgs: reclean-pkgs
	$(MAKE_SELF) stage4-retry-pkgs OPTS="$(OPTS)" PKGS="$(PKGS)"
	@$(call check_pkgs_template,STAGE4_STATUS,LOGOK[[:space:]]\+VFYOK[[:space:]]\+CPOK)
phonytargets += reclean-and-rebuild-pkgs

relook-pkgs:
	$(MAKE_SELF) stage1-retry-pkgs OPTS="-f $(OPTS)" PKGS="$(PKGS)"
	@$(call check_pkgs_template,STAGE1_STATUS,LOGOK)

relook-and-reclean-pkgs: relook-pkgs reclean-pkgs

relook-and-rebuild-pkgs: relook-pkgs reclean-and-rebuild-pkgs

clean-and-verify-noatt: stage1 stage2 stage3 stage4
phonytargets += clean-and-verify-noatt


### Misc

$(SKIPLIST_DIR):
	@mkdir "$(SKIPLIST_DIR)"

$(INCLIST_DIR):
	@mkdir "$(INCLIST_DIR)"

$(TAR_DIR):
	@mkdir "$(TAR_DIR)"

$(SRPM_DIR):
	@mkdir "$(SRPM_DIR)"

$(RPM_DIR):
	@mkdir "$(RPM_DIR)"

$(REBUILD_V_DIR):
	@mkdir "$(REBUILD_V_DIR)"

$(REBUILD_O_DIR):
	@mkdir "$(REBUILD_O_DIR)"

update-versions.sed: $(BIN_LIST)
	sed -e 's,.*[[:space:]]\(\(.\+\)-[^-]\+-[^-]\+\.src\.rpm\)$$,s/^\2-[^-]\\+-[^-]\\+\\.src\\.rpm$$/\1/,' "$(BIN_LIST)" | sort -u >$@
scripts += update-versions.sed

show-srpm-lists:
	@echo "$(SRPM_LISTS)"
phonytargets += show-srpm-lists

update-srpm-versions: update-versions.sed
	sed -i -f $< $(SRPM_LISTS)
phonytargets += update-srpm-versions


### Clean rules

statclean:
	@rm -f $(package_lists) $(srpm_lists) $(scripts) $(ALL_SRPMS) $(skiplists)
phonytargets += statclean

stage1-clean:
	@echo "Really delete Stage I files?"
	@select yn in "Yes" "No"; do case "$$yn" in Yes) rm -rfv "$(STAGE1_STATUS)" "$(STAGE1_LOG)" $(wildcard $(STAGE1_LOG).*) "$(STAGE1_LOGDIR)" "$(SKIPLIST_DIR)" ;; esac; break; done
phonytargets += stage1-clean

stage2-clean:
	@echo "Really delete Stage II files?"
	@select yn in "Yes" "No"; do case "$$yn" in Yes) rm -rfv "$(STAGE2_STATUS)" "$(STAGE2_LOG)" $(wildcard $(STAGE2_LOG).*) "$(STAGE2_LOGDIR)" "$(TAR_DIR)" ;; esac; break; done
phonytargets += stage2-clean

stage3-clean:
	@echo "Really delete Stage III files?"
	@select yn in "Yes" "No"; do case "$$yn" in Yes) rm -rfv "$(STAGE3_STATUS)" "$(STAGE3_LOG)" $(wildcard $(STAGE3_LOG).*) "$(STAGE3_LOGDIR)" "$(SRPM_DIR)" ;; esac; break; done
phonytargets += stage3-clean

stage4-clean:
	@echo "Really delete Stage IV files?"
	@select yn in "Yes" "No"; do case "$$yn" in Yes) rm -rfv "$(STAGE4_STATUS)" "$(STAGE4_LOG)" $(wildcard $(STAGE4_LOG).*) "$(STAGE4_LOGDIR)" "$(RPM_DIR)" ;; esac; break; done
phonytargets += stage4-clean

stage5-clean:
	@echo "Really delete Stage V files?"
	@select yn in "Yes" "No"; do case "$$yn" in Yes) rm -rfv "$(STAGE5_V_STATUS)" "$(STAGE5_V_LOG)" $(wildcard $(STAGE5_V_LOG).*) "$(STAGE5_V_LOGDIR)" "$(REBUILD_V_DIR)" "$(STAGE5_O_STATUS)" "$(STAGE5_O_LOG)" $(wildcard $(STAGE5_O_LOG).*) "$(STAGE5_O_LOGDIR)" "$(REBUILD_O_DIR)" ;; esac; break; done
phonytargets += stage5-clean

clean-indices:
	@rm -fv $(indices)
phonytargets += clean-indices

clean-disttag-index:
	@rm -fv $(DISTTAG_INDEX)
phonytargets += clean-disttag-index


# Self-edit

list-builders:
	@( \
		echo "Host|Arch|Slots|nprocs"; \
		echo "----|----|-----|------"; \
		echo '$(BUILDERS)' | sed -e 's/[[:space:]]\+-H/\n-H/g' | \
			sed -e 's/[[:space:]]*-[dRB][[:space:]]\+"[^"]*"//g' \
				-e 's/[[:space:]]*--[A-Za-z-]\+\(=\+"[^"]*"\)\?//g' \
				-e 's/[[:space:]]*-[A-Za-z][[:space:]]\+"\([^"]*\)"/|\1/g' \
				-e 's/^|//'; \
	) | column -t -s '|'
phonytargets += list-builders

ifneq ($(DEFAULT_NPROFILE),)
ifneq ($(HOST),)
$(HOST).nprofile: | $(DEFAULT_NPROFILE)
	cat "$(DEFAULT_NPROFILE)" >$@
endif
endif

add-builder: NPROCS = $(if $(DEFAULT_NPROFILE),%h.nprofile)
add-builder: $(if $(NOT_CLEANER),,add-cleaner) $(if $(DEFAULT_NPROFILE),$(HOST).nprofile)
	@if [ -z '$(strip $(HOST))' -o -z '$(strip $(ARCH))' -o -z '$(strip $(SLOTS))' -o -z '$(strip $(if $(REPO),$(REPO),$(REPO_DIR)))' ]; \
	then \
		echo 'Usage: add-builder HOST=<host> ARCH=<arch> SLOTS=<slots> REPO=REPO_DIR|<repo> [WORKDIR=<workdir>] [NPROCS=<nprocs>] [NOT_CLEANER=]' >&2; \
		exit 1; \
	fi
	@if echo '$(BUILDERS)' | grep -q -- '-H[[:space:]]\+"$(strip $(HOST))"'; \
	then \
		echo 'ERROR: Builder '\''$(strip $(HOST))'\'' already exists.' >&2; \
		exit 2; \
	fi
	@sed -i -e '/^BUILDERS[[:space:]]*=/,/^\([^B]\|$$\)/ {' \
			-e '/^\([^B]\|$$\)/ i$(backslash)' \
			-e 'BUILDERS += $$(call def_builder,$(strip $(HOST)),$(strip $(ARCH)),$(strip $(SLOTS)),$(if $(strip $(WORKDIR)),$(strip $(WORKDIR)),$(HASHER_WORKDIR)),$(strip $(if $(REPO),$(REPO),$(REPO_DIR)))$(if $(strip $(NPROCS)),$(comma)$(strip $(NPROCS))))' \
			-e '}' $(firstword $(MAKEFILE_LIST))
phonytargets += add-builder

del-builder: DEL_BUILDER = 1
del-builder: $(if $(NOT_CLEANER),,del-cleaner)
	@if [ -z '$(HOST)' ]; then \
		echo 'Usage: del-builder HOST=<host> [NOT_CLEANER=]' >&2; \
		exit 1; \
	fi
	@if [ '$(if $(FORCE),$(FORCE),0)' = '0' ] && \
		! echo '$(BUILDERS)' | grep -q -- '-H[[:space:]]\+"$(strip $(HOST))"'; \
	then \
		echo 'ERROR: Builder '\''$(strip $(HOST))'\'' not found.' >&2; \
		exit 2; \
	fi
	@sed -i -e '/^BUILDERS[[:space:]]*=/,/^\([^B]\|$$\)/ {' \
			-e '/^BUILDERS += $$(call[[:space:]]\+def_builder,$(strip $(HOST))/ d' \
			-e '}' $(firstword $(MAKEFILE_LIST))
phonytargets += del-builder

list-cleaners:
	@( \
		echo "Host|Slots"; \
		echo "----|-----"; \
		echo '$(CLEANERS)' | sed -e 's/[[:space:]]\+-H/\n-H/g' | \
			sed -e 's/[[:space:]]*-[dRB][[:space:]]\+"[^"]*"//g' \
				-e 's/[[:space:]]*-[A-Za-z][[:space:]]\+"\([^"]*\)"/|\1/g' \
				-e 's/^|//'; \
	) | column -t -s '|'
phonytargets += list-cleaners

add-cleaner:
	@if [ -z '$(strip $(HOST))' -o -z '$(strip $(SLOTS))' -o \( -z '$(strip $(SRPMS))' -a -z '$(strip $(if $(REPO),$(REPO),$(REPO_DIR)))' \) ]; \
	then \
		echo 'Usage: add-cleaner HOST=<host> SLOTS=<slots> SRPMS=<srpms> | REPO=REPO_DIR|<repo> [WORKDIR=<workdir>]' >&2; \
		exit 1; \
	fi
	@if echo '$(CLEANERS)' | grep -q -- '-H[[:space:]]\+"$(strip $(HOST))"'; \
	then \
		echo 'ERROR: Cleaner '\''$(strip $(HOST))'\'' already exists.' >&2; \
		exit 2; \
	fi
	@sed -i -e '/^CLEANERS[[:space:]]*=/,/^\([^C]\|$$\)/ {' \
			-e '/^\([^C]\|$$\)/ i$(backslash)' \
			-e 'CLEANERS += $$(call def_cleaner,$(strip $(HOST)),$(strip $(SLOTS)),$(if $(strip $(WORKDIR)),$(strip $(WORKDIR)),$(HASHER_WORKDIR)),$(if $(strip $(SRPMS)),$(strip $(SRPMS)),$(strip $(if $(REPO),$(REPO),$(REPO_DIR)))$(REPO_SRPM_SUBDIR)))' \
			-e '}' $(firstword $(MAKEFILE_LIST))
phonytargets += add-cleaner

del-cleaner: FORCE = $(if $(DEL_BUILDER),$(DEL_BUILDER))
del-cleaner:
	@if [ -z '$(HOST)' ]; then \
		echo 'Usage: del-cleaner HOST=<host>' >&2; \
		exit 1; \
	fi
	@if [ '$(if $(FORCE),$(FORCE),0)' = '0' ] && \
		! echo '$(CLEANERS)' | grep -q -- '-H[[:space:]]\+"$(strip $(HOST))"'; \
	then \
		echo 'ERROR: Cleaner '\''$(strip $(HOST))'\'' not found.' >&2; \
		exit 2; \
	fi
	@sed -i -e '/^CLEANERS[[:space:]]*=/,/^\([^C]\|$$\)/ {' \
			-e '/^CLEANERS += $$(call[[:space:]]\+def_cleaner,$(strip $(HOST))/ d' \
			-e '}' $(firstword $(MAKEFILE_LIST))
phonytargets += del-cleaner

show-repo-dir:
	@echo '$(REPO_DIR)'

set-repo-dir:
	@sed -i -e 's/^REPO_DIR[[:space:]]*=.*$$/REPO_DIR = $(strip $(subst /,\/,$(REPO_DIR)))/' \
		$(firstword $(MAKEFILE_LIST))


# Tail

.SECONDARY: $(package_lists) $(srpm_lists)

.PHONY: $(phonytargets)

.DELETE_ON_ERROR:
